"""
This routine calculates the radar moments for the RPG 94 GHz FMCW radar 'LIMRAD94' and generates a NetCDF4 file.
The generated files can be used as input for the Cloudnet processing chain.

Args:
    **date (string): format YYYYMMDD
    **path (string): path where NetCDF file will be stored

Example:
    python spec2mom_limrad94.py date=20181201 path=/tmp/pycharm_project_626/scripts_Willi/cloudnet_input/

"""

import sys

sys.path.append('../../larda/')
sys.path.append('.')

import numpy as np
from numba import jit
import copy, time
import pyLARDA.helpers as h

from datetime import timedelta



import numpy as np

@jit(nopython=True, fastmath=True)
def estimate_noise_hs74_fast(spectrum, navg=1, std_div=6.0, nnoise_min=1):
    """REFERENCE TO ARM PYART GITHUB REPO: https://github.com/ARM-DOE/pyart/blob/master/pyart/util/hildebrand_sekhon.py

    Estimate noise parameters of a Doppler spectrum.
    Use the method of estimating the noise level in Doppler spectra outlined
    by Hildebrand and Sehkon, 1974.
    Args:
        spectrum (array): Doppler spectrum in linear units.
        navg (int, optional):  The number of spectral bins over which a moving average has been
            taken. Corresponds to the **p** variable from equation 9 of the
            article. The default value of 1 is appropriate when no moving
            average has been applied to the spectrum.
        std_div (float, optional): Number of standard deviations above mean noise floor to specify the
            signal threshold, default: threshold=mean_noise + 6*std(mean_noise)
        nnoise_min (int, optional): Minimum number of noise samples to consider the estimation valid.

    Returns:
        mean (float): Mean of points in the spectrum identified as noise.
        threshold (float): Threshold separating noise from signal. The point in the spectrum with
            this value or below should be considered as noise, above this value
            signal. It is possible that all points in the spectrum are identified
            as noise. If a peak is required for moment calculation then the point
            with this value should be considered as signal.
        var (float): Variance of the points in the spectrum identified as noise.
        nnoise (int): Number of noise points in the spectrum.
        signal_flag (bool): True if spectrum contains a signal
        left_intersec (int): index of intersection of signal and threshold (left side)
        left_intersec (int): index of intersection of signal and threshold (right side)
    References
    ----------
    P. H. Hildebrand and R. S. Sekhon, Objective Determination of the Noise
    Level in Doppler Spectra. Journal of Applied Meteorology, 1974, 13,
    808-811.
    """
    sorted_spectrum = np.sort(spectrum)
    nnoise = len(spectrum)  # default to all points in the spectrum as noise
    n_spec = nnoise

    rtest = 1+1/navg
    sum1 = 0.
    sum2 = 0.
    for i, pwr in enumerate(sorted_spectrum):
        npts = i+1
        if npts < nnoise_min:
            continue

        sum1 += pwr
        sum2 += pwr*pwr

        if npts*sum2 < sum1*sum1*rtest:
            nnoise = npts
        else:
            # partial spectrum no longer has characteristics of white noise.
            sum1 -= pwr
            sum2 -= pwr*pwr
            break

    mean = sum1/nnoise
    var = sum2/nnoise-mean*mean

    threshold = mean + np.sqrt(var) * std_div

    #threshold = sorted_spectrum[nnoise-1]

    # boundaries of major peak only
    left_intersec = -111
    right_intersec = -111

    # save only signal if it as more than 2 points above the noise threshold
    idxMaxSignal = np.argmax(spectrum)
    signal_flag = True

    for ispec in range(idxMaxSignal, n_spec):
        if spectrum[ispec] <= threshold:
            right_intersec = ispec
            break
        else:
            right_intersec = -222  # strong signal till max Nyquist Velocity, maybe folding?

    for ispec in range(idxMaxSignal, -1, -1):
        if spectrum[ispec] <= threshold:
            left_intersec = ispec
            break
        else:
            left_intersec = -222  # strong signal till min Nyquist Velocity, maybe folding?

    return mean, threshold, var, nnoise, signal_flag, left_intersec, right_intersec

@jit(nopython=True, fastmath=True)
def estimate_noise_hs74(spectrum, navg=1.0, std_div=-1.0):
    """
    Estimate noise parameters of a Doppler spectrum.
    Use the method of estimating the noise level in Doppler spectra outlined
    by Hildebrand and Sehkon, 1974.

    References:
        P. H. Hildebrand and R. S. Sekhon, Objective Determination of the Noise
        Level in Doppler Spectra. Journal of Applied Meteorology, 1974, 13, 808-811.

    Args:
        spectrum (numpy.ndarray): dimension (n_Dopplerbins,) Doppler spectrum in linear units.
        **navg (int, optional): The number of spectral bins over which a moving average has been
            taken. Corresponds to the p variable from equation 9 of the
            article.  The default value of 1 is appropriate when no moving
            average has been applied to the spectrum.
        **std_div (float, optional): threshold = number of standard deviations
            above mean noise floor, default threshold is the value of the first
            non-noise value

    Returns:
        list containing

            - mean (float): Mean of points in the spectrum identified as noise.
            - threshold (float): Threshold separating noise from signal. The point in the spectrum with
              this value or below should be considered as noise, above this value
              signal. It is possible that all points in the spectrum are identified
              as noise.  If a peak is required for moment calculation then the point
              with this value should be considered as signal.
            - var (float): Variance of the points in the spectrum identified as noise.
            - nnoise (int): Number of noise points in the spectrum.
    """

    n_spec = len(spectrum)
    sorted_spectrum = np.sort(spectrum)
    nnoise = n_spec  # default to all points in the spectrum as noise
    for npts in range(1, n_spec + 1):
        partial = sorted_spectrum[:npts]
        mean = np.mean(partial)
        var = np.var(partial)
        if var * navg < mean * mean:
            nnoise = npts
        else:
            # partial spectrum no longer has characteristics of white noise
            break

    noise_spectrum = sorted_spectrum[:nnoise]
    mean = np.mean(noise_spectrum)
    var = np.var(noise_spectrum)

    threshold = mean + np.sqrt(var) * std_div

    # boundaries of major peak only
    left_intersec = -111
    right_intersec = -111
    signal_flag = False

    # save only signal if it as more than 2 points above the noise threshold
    # condition = n_spec - nnoise > 0
    condition = True
    # condition = nnoise < n_spec:

    if condition:
        idxMaxSignal = np.argmax(spectrum)
        signal_flag = True

        for ispec in range(idxMaxSignal, n_spec):
            if spectrum[ispec] <= threshold:
                right_intersec = ispec
                break
            else:
                right_intersec = -222  # strong signal till max Nyquist Velocity, maybe folding?

        for ispec in range(idxMaxSignal, -1, -1):
            if spectrum[ispec] <= threshold:
                left_intersec = ispec
                break
            else:
                left_intersec = -222  # strong signal till min Nyquist Velocity, maybe folding?

    return mean, threshold, var, nnoise, signal_flag, left_intersec, right_intersec


@jit(nopython=True)
def find_main_peak(signal):
    idxMaxSignal = np.argmax(signal)
    thresh = np.min(signal)
    len_sig = len(signal)

    for ispec in range(idxMaxSignal, len_sig):
        if signal[ispec] < thresh:
            right_intersec = ispec
            break
        else:
            # strong signal till max Nyquist Velocity, maybe folding?
            right_intersec = -222

    for ispec in range(idxMaxSignal, -1, -1):
        if signal[ispec] < thresh:
            left_intersec = ispec
            # strong signal till max Nyquist Velocity, maybe folding?
            break
        else:
            left_intersec = -222

    return thresh, left_intersec, right_intersec


#
def noise_estimation(data, **kwargs):
    """
    Creates a dict containing the noise threshold, mean noise level,
    the variance of the noise, the number of noise values in the spectrum,
    and the boundaries of the main signal peak, if there is one

    Args:
        data (list): list of data container, containing data[ichirp]['var'] of dimension (n_ts, n_range, n_Doppler_bins)
        **n_std_deviations (float): threshold = number of standard deviations
                                    above mean noise floor, default: threshold is the value of the first
                                    non-noise value

    Returns:
        dict with noise floor estimation for all time and range points
    """

    n_std = kwargs['n_std_deviations'] if 'n_std_deviations' in kwargs else 1.0
    include_noise = kwargs['include_noise'] if 'include_noise' in kwargs else False
    main_peak = kwargs['main_peak'] if 'main_peak' in kwargs else False

    n_chirps = len(data)
    n_dbin = data[0]['var'].shape[2] // 2

    # fill values needs to be masked for noise removal otherwise wrong results
    for ic in range(n_chirps):
        if -999.0 in data[ic]['var']:
            data[ic]['var'][data[ic]['var'] == -999.0] = np.nan

    # if one wants to calculate the moments including the noise
    noise_est = []
    if not include_noise:
        # Estimate Noise Floor for all chirps, time stemps and range gates aka. for all pixels
        # Algorithm used: Hildebrand & Sekhon
        for ic in range(n_chirps):

            n_t = data[ic]['ts'].size
            n_r = data[ic]['rg'].size
            tstart = time.time()

            # allocate numpy arrays
            noise_est.append({'mean': np.zeros((n_t, n_r), dtype=np.float32),
                              'threshold': np.zeros((n_t, n_r), dtype=np.float32),
                              'variance': np.zeros((n_t, n_r), dtype=np.float32),
                              'numnoise': np.zeros((n_t, n_r), dtype=np.int32),
                              'signal': np.full((n_t, n_r), fill_value=True),
                              'bounds': np.full((n_t, n_r, 2), fill_value=None)})

            # gather noise level etc. for all chirps, range gates and times
            for iT in range(n_t):
                # it is ok to check only the first range gate here, because every signal contains at least one value below the noise floor
                nonoise = any(np.isnan(data[ic]['var'][iT, 0, :]))
                for iR in range(n_r):
                    if not nonoise:
                        mean, thresh, var, nnoise, signal, left, right = \
                            estimate_noise_hs74_fast(data[ic]['var'][iT, iR, :], navg=data[ic]['no_av'], std_div=n_std)

                        noise_est[ic]['mean'][iT, iR] = mean
                        noise_est[ic]['variance'][iT, iR] = var
                        noise_est[ic]['numnoise'][iT, iR] = nnoise
                        noise_est[ic]['threshold'][iT, iR] = thresh
                        noise_est[ic]['signal'][iT, iR] = signal
                        noise_est[ic]['bounds'][iT, iR, :] = [left, right]
                        #data[ic]['var'][iT, iR, data[ic]['var'][iT, iR, :] < thresh] = -999.0


                    elif main_peak and not nonoise:

                        # filter for c1 ghost echos?
                        # sum_SaN_neg = np.count_nonzero(~np.isnan(data[ic]['var'][iT, iR, :n_dbin]))
                        # sum_SaN_pos = np.count_nonzero(~np.isnan(data[ic]['var'][iT, iR, n_dbin:]))
                        # if sum_SaN_neg < 10 and sum_SaN_pos < 8 and ic == 0:
                        #    thresh, left, right = np.min(data[ic]['var'][iT, iR, :]), -222, -222
                        #    data[ic]['var'][iT, iR, :] = np.nan
                        # else:
                        thresh, left, right = find_main_peak(data[ic]['var'][iT, iR, :])

                        noise_est[ic]['threshold'][iT, iR] = thresh
                        noise_est[ic]['bounds'][iT, iR, :] = [left, right]

            noise_est[ic]['bounds'][noise_est[ic]['bounds'] == -222] = None
            print('noise removed, chirp = {}, elapsed time = {:.3f} sec.'.format(ic + 1, time.time() - tstart))

    return noise_est


def spectra_to_moments_rpgfmcw94(spectrum_container, **kwargs):
    """
    Calculation of radar moments: reflectivity, mean Doppler velocity, spectral width, skewness, and kurtosis
    translated from Heike's Matlab function
    determination of radar moments of Doppler spectrum over range of Doppler velocity bins

    Note:
        Each chirp of LIMRAD94 data has to be provided separately because
        chirps have in general different n_Doppler_bins and no_av

    Args:
        - spectrum_container (dict): container including VSpec of rpg radar + other variables
        - noise_est (dict): container including mean noise level, noise threshold, ... from noise_estimation routine
        - **include_noise (logical): calculate moments without removing the noise

    Returns:
        dict containing
            - Ze_lin: reflectivity (0.Mom) over range of velocity bins lb to ub [mm6/m3]
            - VEL: mean velocity (1.Mom) over range of velocity bins lb to ub [m/s]
            - sw: spectrum width (2.Mom) over range of velocity bins lb to ub  [m/s]
            - skew: skewness (3.Mom) over range of velocity bins lb to ub
            - kurt: kurtosis (4.Mom) over range of velocity bins lb to ub
    """

    # contains the dimensionality of the Doppler spectrum, (nTime, nRange, nDopplerbins)
    noise_params = ['mean', 'threshold', 'variance', 'signal', 'numnoise', 'bounds']
    n_chirps = len(spectrum_container)
    no_times = spectrum_container[0]['ts'].size
    cum_rg = [0]
    noise_est = []
    no_ranges_tot = 0
    for ic, ichirp in enumerate(spectrum_container):
        cum_rg.append(cum_rg[ic] + ichirp['rg'].size)
        noise_est.append({ivar: spectrum_container[ic][ivar] for ivar in noise_params})
        no_ranges_tot += ichirp['rg'].size

    include_noise = kwargs['include_noise'] if 'include_noise' in kwargs else False
    main_peak = kwargs['main_peak'] if 'main_peak' in kwargs else False

    # initialize variables:
    moments = {'Ze': np.full((no_ranges_tot, no_times), np.nan),
               'VEL': np.full((no_ranges_tot, no_times), np.nan),
               'sw': np.full((no_ranges_tot, no_times), np.nan),
               'skew': np.full((no_ranges_tot, no_times), np.nan),
               'kurt': np.full((no_ranges_tot, no_times), np.nan),
               'mask': np.full((no_ranges_tot, no_times), True)}

    for ic in range(n_chirps):
        tstart = time.time()
        no_ranges = spectrum_container[ic]['rg'].size

        spectra_linear_units = spectrum_container[ic]['var']
        velocity_bins = spectrum_container[ic]['vel']
        DoppRes = spectrum_container[ic]['DoppRes']

        # if ic == 3:
        #    mask = spectra_linear_units < h.lin2z(25.0) and
        #    spectra_linear_units[] = spectra_linear_units

        for iR in range(no_ranges):  # range dimension
            for iT in range(no_times):  # time dimension
                if include_noise:
                    # type1: calculate moments just for main peak, bounded by 2 values lb and ub
                    signal = spectra_linear_units[iT, iR, :]  # extract power spectra in chosen range
                    velocity_bins_extr = velocity_bins  # extract velocity bins in chosen Vdop bin range
                    Ze_lin, VEL, sw, skew, kurt = moment_calculation(signal, velocity_bins_extr, DoppRes)

                else:
                    Ze_lin, VEL, sw, skew, kurt = [np.nan] * 5

                    if noise_est[ic]['signal'][iT, iR]:
                        if main_peak:
                            lb, rb = noise_est[ic]['bounds'][iT, iR, :]
                            spec_no_noise = copy.deepcopy(spectra_linear_units[iT, iR, lb:rb])
                            velocity_bins_extr = velocity_bins[lb:rb]
                            Ze_lin, VEL, sw, skew, kurt = moment_calculation(spec_no_noise, velocity_bins_extr, DoppRes)
                        else:
                            spec_no_noise = copy.deepcopy(spectra_linear_units[iT, iR, :])
                            spec_no_noise[spectra_linear_units[iT, iR, :] < noise_est[ic]['threshold'][iT, iR]] = np.nan
                            Ze_lin, VEL, sw, skew, kurt = moment_calculation(spec_no_noise, velocity_bins, DoppRes)

                iR_tot = cum_rg[ic] + iR

                moments['Ze'][iR_tot, iT] = Ze_lin  # copy temporary Ze_linear variable to output variable
                moments['VEL'][iR_tot, iT] = VEL
                moments['sw'][iR_tot, iT] = sw
                moments['skew'][iR_tot, iT] = skew
                moments['kurt'][iR_tot, iT] = kurt

        print('moments calculated, chrip = {}, elapsed time = {:.3f} sec.'.format(ic + 1, time.time() - tstart))

    # mask all invalid values (NaN)
    for imom in ['Ze', 'VEL', 'sw', 'skew', 'kurt']:
        moments[imom] = np.ma.masked_invalid(moments[imom])

    # mask values <= 0.0
    moments['Ze'] = np.ma.masked_less_equal(moments['Ze'], 0.0)

    # create the mask where invalid values have been encountered
    moments['mask'][np.where(moments['Ze'] > 0.0)] = False

    return moments


@jit(nopython=True)
def check_signal(spec, thresh):
    """
    This helper function checks if a spectrum contains num_cons consecutive values above the noise threshold.

    Args:
        - spec (numpy array, float): contains one spectrum

    Return:
        - sig (logical): True if spectrum contains num_cons consecutive values above the noise threshold, False otherwise.
    """
    num_cons = 6

    for i in range(len(spec) - num_cons - 1):
        sig = True
        for val in spec[i:i + num_cons]:
            if val < thresh or np.isnan(val):
                sig = False
                break

        if sig:
            return True

    return False


@jit(nopython=True, fastmath=True)
def moment_calculation(signal, vel_bins, DoppRes):
    """
    Calculation of radar moments: reflectivity, mean Doppler velocity, spectral width,
        skewness, and kurtosis of one Doppler spectrum. Optimized for the use of Numba.

    Note:
        Divide the signal_sum by 2 because vertical and horizontal channel are added.
        Subtract half of of the Doppler resolution from mean Doppler velocity, because

    Args:
        - signal (float array): detected signal from a Doppler spectrum
        - vel_bins (float array): extracted velocity bins of the signal (same length as signal)
        - DoppRes (int): resolution of the Doppler spectra (different for each chirp)

    Returns:
        dict containing

            - Ze_lin (float array): reflectivity (0.Mom) over range of velocity bins [mm6/m3]
            - VEL (float array): mean velocity (1.Mom) over range of velocity bins [m/s]
            - sw (float array):: spectrum width (2.Mom) over range of velocity bins [m/s]
            - skew (float array):: skewness (3.Mom) over range of velocity bins
            - kurt (float array):: kurtosis (4.Mom) over range of velocity bins
    """

    signal_sum = np.nansum(signal)  # linear full spectrum Ze [mm^6/m^3], scalar
    Ze_lin = signal_sum / 2.0
    pwr_nrm = signal / signal_sum  # determine normalized power (NOT normalized by Vdop bins)

    VEL, sw, skew, kurt = [np.nan] * 4
    if not signal_sum == 0.0:
        VEL = np.nansum(vel_bins * pwr_nrm)
        vel_diff = vel_bins - VEL
        vel_diff2 = vel_diff*vel_diff
        sw = np.sqrt(np.abs(np.nansum(pwr_nrm * vel_diff2)))
        sw2 = sw*sw
        skew = np.nansum(pwr_nrm * vel_diff  * vel_diff2 / (sw * sw2))
        kurt = np.nansum(pwr_nrm * vel_diff2 * vel_diff2 / (sw2 * sw2))
        VEL = VEL - DoppRes / 2.0

    return Ze_lin, VEL, sw, skew, kurt


def filter_ghost_echos_RPG94GHz_FMCW(data, **kwargs):
    ######################################################################
    #
    # 2nd and 3rd chirp ghost echo filter
    if 'clean_spectra' in kwargs and kwargs['clean_spectra']:

        sensitivity_limit = kwargs['SL'] if 'SL' in kwargs else sys.exit(
            'Error in clean_spectra ghost echo filter :: Sensitivity Limit missing!')
        Ze = kwargs['Ze'] if 'Ze' in kwargs else sys.exit(
            'Error in clean_spectra ghost echo filter :: Ze values missing!')

        for ichirp in [0, 1, 2]:

            # threholds for 3rd chrip ghost echo filter
            new_ny_vel = data[ichirp]['vel'].max() - 2.5

            ic_Ze_max = h.z2lin(-22.5)

            idx_left = np.argwhere(-new_ny_vel > data[ichirp]['vel']).max()
            idx_right = np.argwhere(new_ny_vel < data[ichirp]['vel']).min()

            Ze_lin_left = data[ichirp]['var'][:, :, :idx_left].copy()
            Ze_lin_right = data[ichirp]['var'][:, :, idx_right:].copy()

            # if noise was already removed by the RPG software, replace the ghost with -999.,
            # if noise factor 0 was selected in the RPG software, replace the ghost by the minimum spectrum value,
            # to avoid wrong noise estimations (to much signal would be lost otherwise),
            idx_ts_nf0 = np.argwhere(data[ichirp]['var'][:, 0, 0] != -999.0)

            if idx_ts_nf0.size > 0:
                mask_left, mask_right = Ze_lin_left < ic_Ze_max, Ze_lin_right < ic_Ze_max
                min_left, min_right = np.amin(Ze_lin_left, axis=2), np.amin(Ze_lin_right, axis=2)

                for i_bin in range(mask_left.shape[2]):
                    Ze_lin_left[mask_left[:, :, i_bin], i_bin] = min_left[mask_left[:, :, i_bin]]
                for i_bin in range(mask_right.shape[2]):
                    Ze_lin_right[mask_right[:, :, i_bin], i_bin] = min_right[mask_right[:, :, i_bin]]
            else:
                Ze_lin_left[Ze_lin_left < ic_Ze_max] = -999.0
                Ze_lin_right[Ze_lin_right < ic_Ze_max] = -999.0

            data[ichirp]['var'][:, :, :idx_left] = Ze_lin_left.copy()
            data[ichirp]['var'][:, :, idx_right:] = Ze_lin_right.copy()
    ######################################################################
    #
    # 2nd and 3rd chirp ghost echo filter
    if 'C2C3' in kwargs and kwargs['C2C3']:
        for ichirp in [0, 1, 2]:

            # threholds for 3rd chrip ghost echo filter
            new_ny_vel = data[ichirp]['vel'].max() - 2.5

            ic_Ze_max = h.z2lin(-22.5)

            idx_left = np.argwhere(-new_ny_vel > data[ichirp]['vel']).max()
            idx_right = np.argwhere(new_ny_vel < data[ichirp]['vel']).min()

            Ze_lin_left = data[ichirp]['var'][:, :, :idx_left].copy()
            Ze_lin_right = data[ichirp]['var'][:, :, idx_right:].copy()

            # if noise was already removed by the RPG software, replace the ghost with -999.,
            # if noise factor 0 was selected in the RPG software, replace the ghost by the minimum spectrum value,
            # to avoid wrong noise estimations (to much signal would be lost otherwise),
            idx_ts_nf0 = np.argwhere(data[ichirp]['var'][:, 0, 0] != -999.0)

            if idx_ts_nf0.size > 0:
                mask_left, mask_right = Ze_lin_left < ic_Ze_max, Ze_lin_right < ic_Ze_max
                min_left, min_right = np.amin(Ze_lin_left, axis=2), np.amin(Ze_lin_right, axis=2)

                for i_bin in range(mask_left.shape[2]):
                    Ze_lin_left[mask_left[:, :, i_bin], i_bin] = min_left[mask_left[:, :, i_bin]]
                for i_bin in range(mask_right.shape[2]):
                    Ze_lin_right[mask_right[:, :, i_bin], i_bin] = min_right[mask_right[:, :, i_bin]]
            else:
                Ze_lin_left[Ze_lin_left < ic_Ze_max] = -999.0
                Ze_lin_right[Ze_lin_right < ic_Ze_max] = -999.0

            data[ichirp]['var'][:, :, :idx_left] = Ze_lin_left.copy()
            data[ichirp]['var'][:, :, idx_right:] = Ze_lin_right.copy()

    ######################################################################
    #
    # 1st chirp ghost echo filter
    if 'C1' in kwargs and kwargs['C1']:
        # invalid mask has to be provided via kwarg for this filter
        invalid_mask = kwargs['inv_mask']
        rg_offsets = kwargs['offset']
        SL = kwargs['SL']

        # setting higher threshold if chirp 2 contains high reflectivity values
        # sum_over_heightC1 = np.ma.sum(data['Ze'][:rg_offsets[1], :], axis=0)
        sum_over_heightC2 = np.ma.sum(data['Ze'][rg_offsets[1]:rg_offsets[2], :], axis=0)

        # load sensitivity limits (time, height) and calculate the mean over time
        sens_reduction = 15.0  # sensitivity in chirp 1 is reduced by 12.0 dBZ
        C2_Ze_threshold = 18.0  # if sum(C2_Ze) > this threshold, ghost echo in C1 is assumed

        sens_lim = h.z2lin(h.lin2z(np.mean(SL, axis=0)) + sens_reduction)
        ts_to_mask = np.argwhere(h.lin2z(sum_over_heightC2) > C2_Ze_threshold)[:, 0]

        m1 = invalid_mask[:rg_offsets[1], :].copy()
        for idx_ts in ts_to_mask:
            m1[:, idx_ts] = data['Ze'][:rg_offsets[1], idx_ts] < sens_lim
            # a = data['Ze'][:rg_offsets[1], idx_ts] < sens_lim
            # b = data['VEL'][:rg_offsets[1], idx_ts] < -0.5
            # m1[:, idx_ts] = a == b

        invalid_mask[:rg_offsets[1], :] = m1.copy()

        for mom in ['Ze', 'VEL', 'sw', 'skew', 'kurt']:
            data[mom][:rg_offsets[1], :] = np.ma.masked_where(m1, data[mom][:rg_offsets[1], :])

        return invalid_mask


@jit(nopython=True, fastmath=True)
def despeckle(mask, min_percentage):
    """
    SPECKLEFILTER:
        Remove small patches (speckle) from any given mask by checking 5x5 box
        around each pixel, more than half of the points in the box need to be 1
        to keep the 1 at current pixel

    Args:
        mask (numpy.array, integer): mask where 1 = an invalid/fill value and 0 = a data point [height x time]
        min_percentage (float): minimum percentage of neighbours that need to be signal above noise

    Return:
        mask ... speckle-filtered matrix of 0 and 1 that represents (cloud) mask [height x time]

    """

    WSIZE = 5  # 5x5 window
    n_bins = WSIZE*WSIZE
    min_bins = int(min_percentage/100 * n_bins)
    shift = int(WSIZE / 2)
    n_rg, n_ts = mask.shape

    for iR in range(n_rg - WSIZE):
        for iT in range(n_ts - WSIZE):
            if mask[iR, iT] == 1 and np.sum(mask[iR:iR + WSIZE, iT:iT + WSIZE]) > min_bins:
                mask[iR + shift, iT + shift] = 1

    return mask

@jit(nopython=True, fastmath=True)
def despeckle3d(mask, min_percentage):
    """
    SPECKLEFILTER

    last modification: Heike Kalesse, April 26, 2017; kalesse@tropos.de

    TBD:
    - define percentage of neighboring points that have to be 1 in order to keep pxl value as 1 (instead of "hard" number)
    - what is "C" (in output?)

    functionality:
        remove small patches (speckle) from any given mask by checking 5x5 box
        around each pixel, more than half of the points in the box need to be 1
        to keep the 1 at current pixel

    Args:
        mask         ... mask where 1 = an invalid/fill value and 0 = a data point [height x time]
        nr_neighbors ... number of neighbors of pixel that have to be 1 in order to keep pixel value as 1


    Return:
        mask2 ... speckle-filtered matrix of 0 and 1 that represents (cloud) mask [height x time]

    example of a proggi using this function:
    % % % filter out speckles of liq (this is done later in the Shupe 2007 algorithm)
    % % nr_neighbors = 15;  % number of neighbors of pixel (in a 5x5 matrix; i.e., 25pxl) that have to be 1 in order to keep pixel value as 1 in "speckleFilter.m" (orig=12)
    % % [liq_mask,C] = speckleFilter(liq_mask, nr_neighbors);
    20 neighbors in 5x5 matrix means 80%
    """

    window_size = 5  # 5x5x5 window
    n_bins = window_size*window_size*window_size
    min_bins = int(min_percentage/100 * n_bins)
    shift = int(window_size / 2)
    n_rg, n_ts, n_vel = np.array(mask.shape) - window_size

    for iR in range(n_rg):
        for iT in range(n_ts):
            for iB in range(n_vel):
                if mask[iR, iT, iB] == 1:
                    # selecting window of 5x5 pixels
                    window = mask[iR:iR + window_size, iT:iT + window_size, iB:iB + window_size]

                    # if more than n_neighbours pixel in the window are fill values, remove the pixel in the middle
                    if np.sum(window) > min_bins:
                        mask[iR + shift, iT + shift, iB + shift] = 1

    return mask

def make_container_from_spectra(spectra_all_chirps, values, paraminfo, invalid_mask):
    """
    This routine will generate a larda container from calculated moments from spectra.

    Args:
        spectra_all_chirps (list of dicts): dimension [nchirps], containing the spectrum
                                            values of the 94 GHz RPG cloud radar
        values (numpy array): dimension [nrange, ntimes], values of calculated moments
        paraminfo (dict): information from params_[campaign].toml for the specific variable

    Return:
        container (dict): larda data container
    """

    spectra = spectra_all_chirps[0]

    container = {'dimlabel': ['time', 'range'], 'filename': spectra['filename'], 'paraminfo': copy.deepcopy(paraminfo),
                 'rg_unit': paraminfo['rg_unit'], 'colormap': paraminfo['colormap'],
                 'var_unit': paraminfo['var_unit'],
                 'var_lims': paraminfo['var_lims'],
                 'system': paraminfo['system'], 'name': paraminfo['paramkey'],
                 'rg': np.array([rg for ic in spectra_all_chirps for rg in ic['rg']]), 'ts': spectra['ts'],
                 'mask': invalid_mask, 'var': values[:]}

    return container


def build_extended_container(larda, spectra_ch, time_span, **kwargs):
    """
    This routine will generate a list of larda containers including spectra of the RPG-FMCW 94GHz radar.
    The list-container at return will contain the additional information, for each chirp:
        - spec[i_chirps]['no_av'] (float): Number of spectral averages divided by the number of FFT points
        - spec[i_chirps]['DoppRes'] (float): Doppler resolution for
        - spec[i_chirps]['SL'] (2D-float): Sensitivity limit (dimensions: time, range)
        - spec[i_chirps]['NF'] (string): Noise factor, default = 6.0
        - spec[i_chirps]['rg_offsets'] (list): Indices, where chipr shifts

    Args:
        larda (class larda): Initialized pyLARDA, already connected to a specific campaign
        spectra_ch (string): variable name of the spectra to load (in general either "VSpec" or "HSpec")
        time_span (list): Starting and ending time point in datetime format.

    Kwargs:
        **noise_factor (float): Noise factor, number of standard deviations from mean noise floor
        **rm_precip_ghost (bool): Filters ghost echos which occur over all chirps during precipitation.
        **despeckle3d (float): Removes a pixel in "var" if sounding pixels (5x5x5 box) contain at least despeck% nonzeros.
        **estimate_noise (boal): If True, adds the following noise estimation values to the container:
            -   mean (2d ndarray): Mean noise level of the spectra.
            -   threshold (2d ndarray): Noise threshold, values above this threshold are consider as signal.
            -   variance (2d ndarray): The variance of the mean noise level.
            -   numnoise (2d ndarray): Number of Pixels that are cconsideras noise.
            -   signal (2d ndarray): Boolean array, a value is True if no signal was detected.
            -   bounds (3d ndarrax): Dimensions [n_time, n_range, 2] containing the integration boundaries.
    Return:
        container (list): list of larda data container
    """

    # read limrad94 doppler spectra and caluclate radar moments
    std_above_mean_noise = float(kwargs['noise_factor']) if 'noise_factor'    in kwargs else 6.0
    rm_precip_ghost      = kwargs['rm_precip_ghost']     if 'rm_precip_ghost' in kwargs else False
    despeckle3d_perc     = kwargs['do_despeckle3d']      if 'do_despeckle3d'  in kwargs else 80.
    estimate_noise       = kwargs['estimate_noise']      if 'estimate_noise'  in kwargs else False

    AvgNum_in  = larda.read("LIMRAD94", "AvgNum", time_span)
    DoppLen_in = larda.read("LIMRAD94", "DoppLen", time_span)
    MaxVel_in  = larda.read("LIMRAD94", "MaxVel", time_span)

    if spectra_ch[0] == 'H':
        SensitivityLimit = larda.read("LIMRAD94", "SLh", time_span, [0, 'max'])
    else:
        SensitivityLimit = larda.read("LIMRAD94", "SLv", time_span, [0, 'max'])

    # depending on how much files are loaded, AvgNum and DoppLen are multidimensional list
    if len(AvgNum_in['var'].shape) > 1:
        AvgNum = AvgNum_in['var'][0]
        DoppLen = DoppLen_in['var'][0]
        DoppRes = np.divide(2.0 * MaxVel_in['var'][0], DoppLen_in['var'][0])
    else:
        AvgNum = AvgNum_in['var']
        DoppLen = DoppLen_in['var']
        DoppRes = np.divide(2.0 * MaxVel_in['var'], DoppLen_in['var'])

    #  dimensions:
    #       -   LIMRAD_Zspec[:]['var']      [Nchirps][ntime, nrange]
    #       -   LIMRAD_Zspec[:]['vel']      [Nchirps][nDoppBins]
    #       -   LIMRAD_Zspec[:]['no_av']    [Nchirps]
    #       -   LIMRAD_Zspec[:]['DoppRes']  [Nchirps]
    rg_offsets = [0]
    spec = []
    for ic in range(len(AvgNum)):
        tstart = time.time()
        var_string = 'C{}{}'.format(ic + 1, spectra_ch)
        spec.append(larda.read("LIMRAD94", var_string, time_span, [0, 'max']))
        ic_n_ts, ic_n_rg, ic_n_nfft = spec[ic]['var'].shape
        rg_offsets.append(rg_offsets[ic] + ic_n_rg)
        spec[ic].update({'no_av': np.divide(AvgNum[ic], DoppLen[ic]),
                         'DoppRes': DoppRes[ic],
                         'SL': SensitivityLimit['var'][:, rg_offsets[ic]:rg_offsets[ic + 1]],
                         'NF': std_above_mean_noise})
        print('reading C{}{}, elapsed time = {} [hrs:min:sec].'.format(
            ic+1, spectra_ch, timedelta(seconds=int(time.time() - tstart))
        ))

    for ic in range(len(AvgNum)):
        spec[ic]['rg_offsets'] = rg_offsets

    """
    ####################################################################################################################
    ____ ___  ___  _ ___ _ ____ _  _ ____ _       ___  ____ ____ ___  ____ ____ ____ ____ ____ ____ _ _  _ ____ 
    |__| |  \ |  \ |  |  | |  | |\ | |__| |       |__] |__/ |___ |__] |__/ |  | |    |___ [__  [__  | |\ | | __ 
    |  | |__/ |__/ |  |  | |__| | \| |  | |___    |    |  \ |___ |    |  \ |__| |___ |___ ___] ___] | | \| |__]
    
    ####################################################################################################################                                                                                                             
    """

    # 3rd chirp ghost echo filter
    if rm_precip_ghost:
        tstart = time.time()
        filter_ghost_echos_RPG94GHz_FMCW(spec, C2C3=True)
        print('precipitation ghost filter done, elapsed time = {} [hrs:min:sec].'.format(
            timedelta(seconds=int(time.time() - tstart))
        ))

    # despeckle the spectra
    if despeckle3d_perc > 0.0:
        # copy and convert from bool to 0 and 1, remove a pixel if more than 80% of neighbours are invalid (5x5x5 grid)
        for ic in range(len(AvgNum)):
            t0, n_inv0 = time.time(), np.sum(spec[ic]['mask'])
            speckle_mask = despeckle3d(spec[ic]['mask'] * 1, despeckle3d_perc) == 1
            spec[ic]['mask'][speckle_mask] = True
            spec[ic]['var'][speckle_mask]  = -999.0
            t1, n_inv1 = timedelta(seconds=int(time.time() - t0)), np.sum(spec[ic]['mask'])
            print('despeckle3d added {} to {} invalid pixel of the spectra in chirp {}, elapsed time = {} [hrs:min:sec].'.format(
                n_inv1-n_inv0, n_inv0, ic+1, t1
            ))

    # noise estimation a la Hildebrand & Sekhon
    if estimate_noise:
        # Logicals for different tasks
        include_noise = True if spec[0]['NF'] < 0.0 else False
        main_peak = kwargs['main_peak'] if 'main_peak' in kwargs else True

        # remove noise from raw spectra and calculate radar moments
        # dimensions:
        #       -   noise_est[ichirp]['mean']        [Nchirps][ntime, nrange]
        #       -   noise_est[ichirp]['threshold']   [Nchirps][ntime, nrange]
        #       -   noise_est[ichirp]['variance']    [Nchirps][ntime, nrange]
        #       -   noise_est[ichirp]['numnoise']    [Nchirps][ntime, nrange]
        #       -   noise_est[ichirp]['signal']      [Nchirps][ntime, nrange]
        #       -   noise_est[ichirp]['bounds']      [Nchirps][ntime, nrange, 2]
        noise_est = noise_estimation(spec, n_std_deviations=spec[0]['NF'], include_noise=include_noise, main_peak=main_peak)

        # save noise estimation (mean noise, noise threshold to spectra dict
        for iC in range(len(spec)):
            spec[iC].update({key: noise_est[iC][key].copy() for key in noise_est[iC].keys()})

    return spec


def spectra2moments(Z_spec, paraminfo, **kwargs):
    """
    This routine calculates the radar moments: reflectivity, mean Doppler velocity, spectrum width, skewness and
    kurtosis from the level 0 spectrum files of the 94 GHz RPG cloud radar.

    Args:
        Z_spec (list of dicts): list containing the dicts for each chrip of RPG-FMCW Doppler cloud radar
        rg_offsets (list): containing indices of chrip change
        paraminfo (dict): information from params_[campaign].toml for the system LIMRAD94
        SL (dict): larda container of sensitivity limit
        nsd (float): number of standard deviations above mean noise threshold

    Return:
        container_dict (dict): dictionary of larda containers, including larda container for Ze, VEL, sw, skew, kurt

    """

    ####################################################################################################################
    #
    # moment calculation
    include_noise = True if Z_spec[0]['NF'] < 0.0 else False
    main_peak = kwargs['main_peak'] if 'main_peak' in kwargs else True
    moments = spectra_to_moments_rpgfmcw94(Z_spec, include_noise=include_noise, main_peak=main_peak)
    invalid_mask = moments['mask'].copy()

    ####################################################################################################################
    #
    # 1st chirp ghost echo filter
    # test differential phase filter technique
    if 'filter_ghost_C1' in kwargs and kwargs['filter_ghost_C1']:
        invalid_mask = filter_ghost_echos_RPG94GHz_FMCW(moments,
                                                        C1=True,
                                                        inv_mask=invalid_mask,
                                                        offset=Z_spec[0]['rg_offsets'],
                                                        SL=Z_spec[0]['SL'])

    ####################################################################################################################
    #
    # despeckle the moments
    if 'despeckle' in kwargs and kwargs['despeckle']:
        tstart = time.time()
        # copy and convert from bool to 0 and 1, remove a pixel  if more than 20 neighbours are invalid (5x5 grid)
        new_mask = despeckle(invalid_mask.copy() * 1, 80.)
        invalid_mask[new_mask == 1] = True

    for mom in ['Ze', 'VEL', 'sw', 'skew', 'kurt']:
        moments[mom][invalid_mask] = -999.0

        print('despeckle done, elapsed time = {:.3f} sec.'.format(time.time() - tstart))

    ####################################################################################################################
    #
    # build larda containers from calculated moments
    container_dict = {mom: make_container_from_spectra(Z_spec, moments[mom].T, paraminfo[mom], invalid_mask.T) for mom in ['Ze', 'VEL', 'sw', 'skew', 'kurt']}

    return container_dict



#!/usr/bin/env python3

from prompt_toolkit.application import Application
from prompt_toolkit.document import Document
from prompt_toolkit.filters import has_focus
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import TextArea
from prompt_toolkit.completion import PathCompleter


import sys, traceback
sys.path.append('../')
import datetime
# just needed to find pyLARDA from this location
# sys.path.append('../larda/')
# sys.path.append('.')
import pyLARDA
import pyLARDA.helpers as h
import pyLARDA.Connector as Connector

import logging
log = logging.getLogger('__main__')
log.setLevel(logging.DEBUG)
# write logs to file for debugging reasons
#log.addHandler(logging.FileHandler("promt_log", mode='w', encoding=None, delay=False))

help_text = """
Press Ctrl-f and Ctrl-b to toggle fields, ENTER to update.
Tab autocompletes the base dir and arrow keys acccess the history.
Ctrl-C will exit and provide the final output in the terminal.
"""

system_info = {
    'path': {
        'nc': 
            {'base_dir': '/lacroshome/mira/data/', 
            'matching_subdirs': '(Mom\/\d{4}.*\d{8}_\d{6}.mmclx)', 
            'date_in_filename': '(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})_(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})'}},
    'generic': {}, 
    'params': {}}


def get_the_filelist(system_info):

    log.info(system_info['path']['nc']['base_dir'])
    log.info(system_info['path']['nc']['matching_subdirs'])
    log.info(system_info['path']['nc']['date_in_filename'])

    conn = Connector.Connector('empty', system_info, [["20140101", datetime.datetime.utcnow().strftime("%Y%m%d")]])
    conn.build_filehandler()

    return conn.filehandler


def format_files(d):

    s = ''

    for k,v in d.items():
        s += '{}\n'.format(k)
        s += '\n'.join([str(e) for e in v])
    return s



def main():

    output_field = TextArea(style="class:output-field", text=help_text)
    dir_field = TextArea(
        height=1,
        prompt="base_dir ",
        text=system_info['path']['nc']['base_dir'],
        style="class:input-field",
        multiline=False,
        
        completer=PathCompleter(),
        wrap_lines=False,
    )
    matching_field = TextArea(
        height=1,
        prompt="matching_subdirs ",
        text=system_info['path']['nc']['matching_subdirs'],
        style="class:input-field",
        multiline=False,
        wrap_lines=False,
    )
    date_file_field = TextArea(
        height=1,
        prompt="date_in_filename ",
        text=system_info['path']['nc']['date_in_filename'],
        style="class:input-field",
        multiline=False,
        wrap_lines=False,
    )

    container = HSplit(
        [
            dir_field,
            matching_field,
            date_file_field,
            Window(height=1, char="-", style="class:line"),
            output_field,
        ]
    )

    # The key bindings.
    kb = KeyBindings()

    kb.add("c-f")(focus_next)
    kb.add("c-b")(focus_previous)

    @kb.add("c-c")
    @kb.add("c-q")
    def _(event):
        " Pressing Ctrl-Q or Ctrl-C will exit the user interface. "
        event.app.exit()


    #@kb.add('enter')
    # Attach accept handler to the input field. We do this by assigning the
    # handler to the `TextArea` that we created earlier. it is also possible to
    # pass it to the constructor of `TextArea`.
    # NOTE: It's better to assign an `accept_handler`, rather then adding a
    #       custom ENTER key binding. This will automatically reset the input
    #       field and add the strings to the history.
    def accept(buff):
        # Evaluate "calculator" expression.
        log.info(str(buff))
        #log.info(type(buff))
        system_info['path']['nc']['base_dir'] = dir_field.text
        system_info['path']['nc']['matching_subdirs'] = matching_field.text
        system_info['path']['nc']['date_in_filename'] = date_file_field.text
        log.info(str(matching_field.__dict__))
        #log.info('matching_field.buffer ' + str(matching_field.buffer.__dict__))
        #log.info('matching_field._working_lines ' + str(matching_field.buffer._working_lines))
        #log.info('history get_strings ' + str(matching_field.buffer.history.get_strings()))


        # hack to make the history work (aka accessing old entries with arrow keys)
        # despite not resetting the field
        dir_field.buffer._working_lines.append(system_info['path']['nc']['base_dir'])
        matching_field.buffer._working_lines.append(system_info['path']['nc']['matching_subdirs'])
        date_file_field.buffer._working_lines.append(system_info['path']['nc']['date_in_filename'])
        # try:
        #     output = "\n\nIn:  {}\nOut: {}".format(
        #         input_field.text, eval(input_field.text)
        #     )  # Don't do 'eval' in real code!
        # except BaseException as e:
        #     output = "\n\n{}".format(e)

        try:
            files = get_the_filelist(system_info)
            new_text = format_files(files)
        except BaseException as e:
            tb = traceback.format_exc()
            new_text = "\n{}\n{}".format(tb, e)

        # Add text to output buffer.
        output_field.buffer.document = Document(
            text=new_text, cursor_position=len(new_text)
        )

        return True

    dir_field.accept_handler = accept
    matching_field.accept_handler = accept
    date_file_field.accept_handler = accept


    # Style.
    style = Style(
        [
            ("output-field", "bg:#000044 #ffffff"),
            ("input-field", "bg:#000000 #ffffff"),
            ("line", "#004400"),
        ]
    )

    # Run application.
    application = Application(
        layout=Layout(container, focused_element=dir_field),
        key_bindings=kb,
        style=style,
        mouse_support=True,
        full_screen=True,
    )

    application.run()
    print('')
    print('[SYSTEM.path.nc]')
    print("  base_dir = '{}'".format(system_info['path']['nc']['base_dir']))
    print("  matching_subdirs = '{}'".format(system_info['path']['nc']['matching_subdirs']))
    print("  date_in_filename = '{}'".format(system_info['path']['nc']['date_in_filename']))
    print('')


if __name__ == "__main__":
    main()



#!/usr/bin/python


import datetime, os, copy
import numpy as np
import pprint as pp
import re
import errno
import ast
import traceback

from functools import reduce

import logging

logger = logging.getLogger(__name__)


def ident(x):
    return x


def get_converter_array(string, **kwargs):
    """colletion of converters that works on arrays
    combines time, range and varconverters (i see no conceptual separation here)
   
    the maskconverter becomes relevant, if the order is no
    time, range, whatever (as in mira spec)

    chaining example:
    ```var_conversion = 'z2lin,extrfromaxis2(0)'```

    Returns:
        (varconverter, maskconverter) which both are functions
    """
    if ',' in string:
        converters = [get_converter_array(s, **kwargs) for s in string.split(',')]
        varfuncs = reversed([f[0] for f in converters])
        maskfuncs = reversed([f[1] for f in converters])
        varf = lambda x: reduce(lambda r, f: f(r), varfuncs, x)
        maskf = lambda x: reduce(lambda r, f: f(r), maskfuncs, x)

        return varf, maskf


    elif string == 'since20010101':
        return lambda x: x + dt_to_ts(datetime.datetime(2001, 1, 1)), ident
    elif string == 'hours_since20150101':
        return lambda x: x*60*60 + dt_to_ts(datetime.datetime(2015, 1, 1)), ident
    elif string == 'unix':
        return lambda x: x, ident
    elif string == 'since19691231':
        return lambda x: x + dt_to_ts(datetime.datetime(1969, 12, 31, 23)), ident
    elif string == 'since19700101':
        return lambda x: x + dt_to_ts(datetime.datetime(1970, 1, 1)), ident
    elif string == 'since19040101':
        return lambda x: x + dt_to_ts(datetime.datetime(1904, 1, 1)), ident
    elif string == 'beginofday':
        if 'ncD' in kwargs.keys():
            return (lambda h: (h.astype(np.float64) * 3600. + \
                               float(dt_to_ts(datetime.datetime(kwargs['ncD'].year,
                                                                kwargs['ncD'].month,
                                                                kwargs['ncD'].day)))),
                    ident)
    elif string == "hours_since_year0":
        return (lambda x: x*24*60*60 - 62167305599.99999,
                ident)
    elif string == "pollytime":
        return (lambda x: np.array([x[i,1] + dt_to_ts(datetime.datetime.strptime(str(int(x[i,0])), "%Y%m%d"))\
                for i in range(x.shape[0])]),
                ident)
    elif string == 'since20200101':
        return lambda x: x + dt_to_ts(datetime.datetime(2020, 1, 1,)), ident


    elif string == "km2m":
        return lambda x: x * 1000., ident
    elif string == "sealevel2range":
        return lambda x: x - kwargs['altitude'], ident

    elif string == 'z2lin':
        return z2lin, ident
    elif string == 'lin2z':
        return lin2z, ident
    elif string == 'switchsign':
        return lambda x: -x, ident

    elif string == "mira_azi_offset":
        return lambda x: (x + kwargs['mira_azi_zero']) % 360, ident

    elif string == 'transposedim':
        #return np.transpose, np.transpose
        return transpose_only, transpose_only
    elif string == 'transposedim+invert3rd':
        return transpose_and_invert, transpose_and_invert
    elif string == 'divideby2':
        return divide_by(2.), ident
    elif string == 'keepNyquist':
        return ident, ident
    elif string == 'raw2Z':
        return raw2Z(**kwargs), ident
    elif string == "extract_level0":
        return lambda x: x[:, 0], ident
    elif string == "extract_level1":
        return lambda x: x[:, 1], ident
    elif string == "extract_level2":
        return lambda x: x[:, 2], ident
    elif string == 'extract_1st':
        return lambda x: np.array(x[0])[np.newaxis,], ident
    elif string == "none":
        return ident, ident
    elif 'extrfromaxis2' in string:
        return get_extrfromaxis2(string), get_extrfromaxis2(string)

    else:
        raise ValueError("converter {} not defined".format(string))


def transpose_only(var):
    return np.transpose(var)[:, :, :]


def transpose_and_invert(var):
    return np.transpose(var)[:, :, ::-1]


def divide_by(val):
    return lambda var: var / val


def flatten(xs):
    """flatten inhomogeneous deep lists
    e.g. ``[[1,2,3],4,5,[6,[7,8],9],10]``
    """
    result = []
    if isinstance(xs, (list, tuple)):
        for x in xs:
            result.extend(flatten(x))
    else:
        result.append(xs)
    return result


def since2001_to_dt(s):
    """seconds since 2001-01-01 to datetime"""
    # return (dt - datetime.datetime(1970, 1, 1)).total_seconds()
    return datetime.datetime(2001, 1, 1) + datetime.timedelta(seconds=s)


def dt_to_ts(dt):
    """datetime to unix timestamp"""
    # return dt.replace(tzinfo=datetime.timezone.utc).timestamp()
    return (dt - datetime.datetime(1970, 1, 1)).total_seconds()


def ts_to_dt(ts):
    """unix timestamp to dt"""
    return datetime.datetime.utcfromtimestamp(ts)


def argnearest(array, value):
    """find the index of the nearest value in a sorted array
    for example time or range axis

    Args:
        array (np.array): sorted array with values, list will be converted to 1D array
        value: value to find
    Returns:
        index  
    """
    if type(array) == list:
        array = np.array(array)
    i = np.searchsorted(array, value) - 1

    if not i == array.shape[0] - 1:
            if np.abs(array[i] - value) > np.abs(array[i + 1] - value):
                i = i + 1
    return i


def nearest(array, pivot):
    """find the nearest value to a given one

    Args:
        array (np.array): sorted array with values
        pivot: value to find
    Returns:
        value with smallest distance
    """
    return min(array, key=lambda x: abs(x - pivot))


def lin2z(array):
    """linear values to dB (for np.array or single number)"""
    return 10 * np.ma.log10(array)


def z2lin(array):
    """dB to linear values (for np.array or single number)"""
    return 10 ** (array / 10.)


def raw2Z(array, **kwargs):
    """raw signal units (MRR-Pro) to reflectivity Z"""
    return array * kwargs['wl']**4 / (np.pi**5) / 0.93 * 10**6


def get_extrfromaxis2(string):
    """get function that extracts given index from axis2"""

    m = re.search(r"\((\d+)\)", string)
    ind = int(m.groups(0)[0])
    return lambda x: x[:,:,ind]


def fill_with(array, mask, fill):
    """fill an array where mask is true with fill value"""
    filled = array.copy()
    filled[mask] = fill
    return filled


def guess_str_to_dict(string):
    """try to convert a text string into a dict
    intended to be used in the var_def

    Returns:
        dict with flag as key and description string
    """


    if "{" in string:
        #probalby already the stringified python format
        return ast.literal_eval(string)

    elif "\nValue" in string:
        # the cloudnetpy format \nValue 0: desc\n ....
        d = {}
        for e in string.split('\nValue '):
            if len(e) > 0:
                k, v = e.split(':')
                d[int(k)] = v.strip()
        return d
    elif "\n" in string:
        # the cloudnet format 0: desc\n ....
        d = {}
        for e in string.split('\n'):
            k, v = e.split(':')
            d[int(k)] = v.strip()
        return d
    elif "\\n" in string:
        # pollynet format
        d = {}
        for e in string.split('\\n'):
            m = re.match(r'(\d{1,2}): (.*)', e.replace(r'\"', ''))
            d[int(m.group(1))] = m.group(2).strip()
        return d
    else:
        # unknown format
        return string


def _method_info_from_argv(argv=None):
    """Command-line -> method call arg processing.

    - positional args:
            a b -> method('a', 'b')
    - intifying args:
            a 123 -> method('a', 123)
    - json loading args:
            a '["pi", 3.14, null]' -> method('a', ['pi', 3.14, None])
    - keyword args:
            a foo=bar -> method('a', foo='bar')
    - using more of the above
            1234 'extras=["r2"]'  -> method(1234, extras=["r2"])

    @param argv {list} Command line arg list. Defaults to `sys.argv`.
    @returns (<method-name>, <args>, <kwargs>)

    Reference: http://code.activestate.com/recipes/577122-transform-command-line-arguments-to-args-and-kwarg/
    """
    import json
    import sys
    if argv is None:
        argv = sys.argv

    method_name, arg_strs = argv[0], argv[1:]
    args = []
    kwargs = {}
    for s in arg_strs:
        if s.count('=') == 1:
            key, value = s.split('=', 1)
        else:
            key, value = None, s
        try:
            value = json.loads(value)
        except ValueError:
            pass
        if key:
            kwargs[key] = value
        else:
            args.append(value)
    return method_name, args, kwargs


def reshape_spectra(data):
    """This function reshapes time, range and var variables of a data container and returns numpy arrays.

    Args:
        data (dict): data container

    Returns:
        list with

        - ts (numpy.array): time stamp numpy array, dim = (n_time,)
        - rg (numpy.array): range stamp numpy array, dim = (n_range,)
        - var (numpy.array): values of the spectra numpy array, dim = (n_time, n_range, n_vel)
    """
    n_ts, n_rg, n_vel = data['ts'].size, data['rg'].size, data['vel'].size

    if data['dimlabel'] == ['time', 'range', 'vel']:
        ts = data['ts'].copy()
        rg = data['rg'].copy()
        var = data['var'].copy()
        mask = data['mask'].copy()
    elif data['dimlabel'] == ['time', 'vel']:
        ts = data['ts'].copy()
        rg = np.reshape(data['rg'], (n_rg,))
        var = np.reshape(data['var'], (n_ts, 1, n_vel))
        mask = np.reshape(data['mask'], (n_ts, 1, n_vel))
    elif data['dimlabel'] == ['range', 'vel']:
        ts = np.reshape(data['ts'].copy(), (n_ts,))
        rg = data['rg'].copy()
        var = np.reshape(data['var'], (1, n_rg, n_vel))
        mask = np.reshape(data['mask'], (1, n_rg, n_vel))
    elif data['dimlabel'] == ['vel']:
        ts = np.reshape(data['ts'].copy(), (n_ts,))
        rg = np.reshape(data['rg'], (n_rg,))
        var = np.reshape(data['var'], (1, 1, n_vel))
        mask = np.reshape(data['mask'], (1, 1, n_vel))
    else:
        raise TypeError('Wrong data format in plot_spectra')

    return ts, rg, var, mask


def pformat(data, verbose=False):
    """return a pretty string from a data_container"""
    string = []
    string.append("== data container: system {} name {}  ==".format(data["system"], data["name"]))
    string.append("dimlabel    {}".format(data["dimlabel"]))
    if "time" in data["dimlabel"]:
        string.append("timestamps  {} {} to {}".format(
            data["ts"].shape,
            ts_to_dt(data["ts"][0]), ts_to_dt(data["ts"][-1])))
    elif "ts" in data.keys():
        string.append("timestamp   {}".format(ts_to_dt(data['ts'])))
    if "range" in data["dimlabel"]:
        string.append("range       {} {:7.2f} to {:7.2f}".format(
            data["rg"].shape,
            data["rg"][0], data["rg"][-1]))
        string.append("rg_unit     {}".format(data["rg_unit"]))
    elif "rg" in data.keys():
        string.append("range       {}".format(data['rg']))
        string.append("rg_unit     {}".format(data["rg_unit"]))
    if "vel" in data.keys():
        string.append("vel         {}  {:5.2f} to {:5.2f}".format(
            data["vel"].shape,
            data["vel"][0], data["vel"][-1]))
    if not np.all(data["mask"]):
        string.append("var         {}  min {:7.2e} max {:7.2e}".format(
            data['var'].shape,
            np.min(data['var'][~data['mask']]), np.max(data['var'][~data['mask']])))
        string.append("            mean {:7.2e} median {:7.2e}".format(
            np.mean(data['var'][~data['mask']]), np.median(data['var'][~data['mask']])))
    string.append("mask        {:4.1f}%".format(
        np.sum(data["mask"])/data['mask'].ravel().shape[0]*100.))
    string.append("var_unit    {}".format(data["var_unit"]))
    string.append("var_lims    {}".format(data["var_lims"]))
    string.append("default colormap {}".format(data["colormap"]))
    if verbose:
        string.append("filenames")
        string.append(pp.pformat(data["filename"], indent=2))
        string.append("paraminfo".format())
        string.append(pp.pformat(data['paraminfo'], indent=2))
    return "\n".join(string)


def isKthBitSet(n, k):
    """
    Function to check if a certain bit of a number is set (required to analyse quality flags)
    """
    if n & (1 << (k - 1)):
        return 1
    else:
        return 0


def pprint(data, verbose=False):
    """print a pretty representation of the data container"""
    print(pformat(data, verbose=verbose))


def extract_case_from_excel_sheet(data_loc, sheet_nr=0, **kwargs):
    """This function extracts information from an excel sheet. It can be used for different scenarios.
    The first row of the excel sheet contains the headline, defined as follows:

    +----+-------+-------+-------+-------+-------+-------+------------+-------+
    |    |   A   |   B   |   C   |   D   |   E   |   F   |      G     |   H   |
    +----+-------+-------+-------+-------+-------+-------+------------+-------+
    |  1 |  date | start |  end  |   h0  |  hend |  MDF  |  noise_fac | notes |
    +----+-------+-------+-------+-------+-------+-------+------------+-------+


                                OR

    +----+------------+----------+-------+-------+-----------+--------+
    |    |      A     |     B    |   C   |   D   |     E     |    F   |
    +----+------------+----------+-------+-------+-----------+--------+
    |  1 |  datestart |  dateend |   h0  |  hend | noise_fac |  notes |
    +----+------------+----------+-------+-------+-----------+--------+


    The following rows contain the cases of interest. Make sure that the ALL the data in the excel sheet is formatted as
    string! The data has to be provided in the following syntax:

        - date (string): format YYYYMMDD
        - start (string): format HHMMSS
        - datestart (string): format YYYYMMDDHHMMSS
        - dateend (string): format YYYYMMDDHHMMSS
        - end (string): format HHMMSS
        - h0 (string): minimum height
        - hend (string): maximum height
        - MDF (string): name of the MDF used for this case
        - noise_fac (string): noise factor
        - notes (string): additional notes for the case (stored but not in use by the program)

    Args:
        data_loc (string): path to the excel file (make sure the data_loc contains the suffix .xlsx)
        sheet_nr (integer): number of the desired excel sheet

    Returns:
        case_list contains the information for all cases
            
        - begin_dt (datetime object): start of the time interval
        - end_dt (datetime object): end of the time interval
        - plot_range (list): height interval
        - MDF_name (string): name of MDF used for this case
        - noisefac (string): number of standard deviations above mean noise level
        - notes (string): additional notes for the user
            
    """

    import xlrd

    excel_sheet = xlrd.open_workbook(data_loc)
    sheet = excel_sheet.sheet_by_index(sheet_nr)
    case_list = []

    if kwargs['kind'] == 'ann_input':
        # exclude header from data
        for icase in range(1, sheet.nrows):
            irow = sheet.row_values(icase)
            irow[:2] = [int(irow[i]) for i in range(2)]

            if irow[5] != 'ex':
                case_list.append({
                    'begin_dt': datetime.datetime.strptime(str(irow[0]), '%Y%m%d%H%M%S'),
                    'end_dt': datetime.datetime.strptime(str(irow[1]), '%Y%m%d%H%M%S'),
                    'plot_range': [float(irow[2]), float(irow[3])],
                    'noise_fac': irow[4],
                    'notes': irow[5]})

    if kwargs['kind'] == 'cumulustest':
        # exclude header from data
        for icase in range(1, sheet.nrows):
            irow = sheet.row_values(icase)
            irow[:3] = [int(irow[i]) for i in range(3)]

            if irow[7] != 'ex':
                case_list.append({
                    'begin_dt': datetime.datetime.strptime(str(irow[0]) + ' ' + str(irow[1]), '%Y%m%d %H%M%S'),
                    'end_dt': datetime.datetime.strptime(str(irow[0]) + ' ' + str(irow[2]), '%Y%m%d %H%M%S'),
                    'plot_range': [float(irow[3]), float(irow[4])],
                    'MDF_name': irow[5],
                    'noisefac': irow[6],
                    'notes': irow[7]})

    return case_list


def interp_only_3rd_dim(arr, old, new):
    """function to interpolate only the velocity (3rd) axis"""

    from scipy import interpolate

    f = interpolate.interp1d(old, arr, axis=2,
                             bounds_error=False, fill_value=-999.)
    new_arr = f(new)

    return new_arr


def put_in_container(data, data_container, **kwargs):
    """
    This routine will generate a new larda container, replacing var in the data_container given with data

    Kwargs:
        all keys used in larda containers, that is
        paraminfo (dict): information from params_[campaign].toml for the specific variable
        rg_unit: range unit
        mask: ndarray of same size as var, indicating masked values
        ts: time stamp in unix time
        var_lims: variable limits
        colormap: colormap to be used for plotting


    Return:
        container (dict): larda data container
    """

    container = copy.deepcopy(data_container)
    container['var'] = data
    container.update(kwargs)

    return container


def change_dir(folder_path, **kwargs):
    """
    This routine changes to a folder or creates it (including subfolders) if it does not exist already.

    Args:
        folder_path (string): path of folder to switch into
    """

    folder_path = folder_path.replace('//', '/', 1)

    if not os.path.exists(os.path.dirname(folder_path)):
        try:
            os.makedirs(os.path.dirname(folder_path))
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    os.chdir(folder_path)
    logger.debug('\ncd to: {}'.format(folder_path))


def make_dir(folder_path):
    """
    This routine changes to a folder or creates it (including subfolders) if it does not exist already.

    Args:
        folder_path (string): path of folder to switch into
    """

    if not os.path.exists(os.path.dirname(folder_path)):
        try:
            os.makedirs(os.path.dirname(folder_path))
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise


def print_traceback(txt):
    """
    Print the traceback to an error to screen.
    Args:
        - txt (string): error msg
    """
    print(txt)
    track = traceback.format_exc()
    print(track)


def smooth(y, box_pts, padding='constant'):
    """Smooth a one dimensional array using a rectangular window of box_pts points

    Args:
        y (np.array): array to be smoothed
        box_pts: number of points of the rectangular smoothing window
    Returns:
        y_smooth (np.arrax): smoothed array
    """

    box = np.ones(box_pts) / box_pts
    if padding.lower() == 'constant':
        return np.convolve(y, box, mode='full')[box_pts // 2:-box_pts // 2 + 1]
    else:
        return np.convolve(y, box, mode='same')




#!/usr/bin/python3
# coding=utf-8
""" """

"""
Author: radenz@tropos.de

"""

import sys
sys.path.append('../')
import pyLARDA
import os
from pathlib import Path
import argparse

import logging
log = logging.getLogger('pyLARDA')
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler())

ROOT_DIR = Path(__file__).absolute().parents[1]
print(ROOT_DIR)

parser = argparse.ArgumentParser(
    description='''
    Example `python3 ListCollector.py -c oceanet_pascal lacros_accept`.
    As default all campaigns are collected.'''
)
parser.add_argument('-c', '--campaign', nargs='+',
                    help='just run for a defined campaign(s)')
args = parser.parse_args()

camp = pyLARDA.LARDA_campaign(ROOT_DIR / Path("../larda-cfg/"), "campaigns.toml")
camp_list = camp.get_campaign_list()
print(camp_list)

if args.campaign:
    assert set(args.campaign).issubset(camp_list), 'campaign not in list'
    camp_list = args.campaign

#larda=pyLARDA.LARDA().connect('test_filepatterns', build_lists=True)

for cname in camp_list:
    larda=pyLARDA.LARDA().connect(cname, build_lists=True)
    print(larda.connectors.keys())
    
#larda=pyLARDA.LARDA('COLRAWI')
#larda=pyLARDA.LARDA('LACROS_at_Leipzig', build_lists=True)



#!/usr/bin/python3

import sys

# just needed to find pyLARDA from this location
sys.path.append('../')
sys.path.append('.')

import matplotlib

matplotlib.use('Agg')

import pyLARDA
import datetime

import logging
log = logging.getLogger('pyLARDA')
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler())

# Load LARDA
larda = pyLARDA.LARDA().connect('lacros_dacapo')

#case_prefix = 'plots/scatter_case_studies/comparison_cloud_'
case_prefix = 'plots/'
c_info = [larda.camp.LOCATION, larda.camp.VALID_DATES]

print(larda.days_with_data())

# begin_dt=datetime.datetime(2018,12,6,0,1)
begin_dt = datetime.datetime(2018, 12, 6, 0, 0, 0)
end_dt   = datetime.datetime(2018, 12, 6, 0, 30, 0)

plot_range = [0, 12000]

# load the reflectivity data
MIRA_Zg = larda.read("MIRA", "Zg", [begin_dt, end_dt], [0, 'max'])
MIRA_Zg['var_lims'] = [-60, 20]

LIMRAD94_Z = larda.read("LIMRAD94", "Ze", [begin_dt, end_dt], [0, 'max'])
LIMRAD94_Z['var_lims'] = [-60, 20]

LIMRAD94_Z_interp = pyLARDA.Transformations.interpolate2d(LIMRAD94_Z, new_time=MIRA_Zg['ts'], new_range=MIRA_Zg['rg'])

fig, ax = pyLARDA.Transformations.plot_scatter(MIRA_Zg, LIMRAD94_Z_interp, var_lim=[-75, 20],
                                          custom_offset_lines=5.0, z_converter='lin2z')

fig.savefig(case_prefix+'scatter_mira_limrad_Z.png', dpi=250)

# load the Doppler velocity data
MIRA_VELg = larda.read("MIRA", "VELg", [begin_dt, end_dt], [0, 'max'])
LIMRAD94_VEL = larda.read("LIMRAD94", "VEL", [begin_dt, end_dt], [0, 'max'])

LIMRAD94_VEL_interp = pyLARDA.Transformations.interpolate2d(LIMRAD94_VEL,
                                                            new_time=MIRA_VELg['ts'], new_range=MIRA_VELg['rg'])

fig, ax = pyLARDA.Transformations.plot_scatter(MIRA_VELg, LIMRAD94_VEL_interp, var_lim=[-6, 4],
                                          custom_offset_lines=1.0)

fig.savefig(case_prefix+'scatter_mira_limrad_VEL.png', dpi=250)


## load the reflectivity data
#MIRA_Zg = larda.read("CLOUDNET", "Z", [begin_dt, end_dt], [0, 'max'])
#MIRA_Zg['var_lims'] = [-60, 20]
#
#fig, ax = pyLARDA.Transformations.plot2d(MIRA_Zg, range_interval=plot_range, z_converter='lin2z')
#fig.savefig(case_prefix+'cloudnet_mira_Z.png', dpi=250)
#
#LIMRAD94_Z = larda.read("CLOUDNET_LIMRAD", "Z", [begin_dt, end_dt], [0, 'max'])
#LIMRAD94_Z['var_lims'] = [-60, 20]
#
#fig, ax = pyLARDA.Transformations.plot2d(LIMRAD94_Z, range_interval=plot_range, z_converter='lin2z')
#fig.savefig(case_prefix+'cloudnet_limrad_Z.png', dpi=250)


#fig, ax = pyLARDA.Transformations.scatter(MIRA_Zg, LIMRAD94_Z, var_lim=[-75, 20],
#                                          custom_offset_lines=5.0, z_converter='lin2z')
#
#fig.savefig(case_prefix+'cloudnet_scatter_mira_limrad_Z.png', dpi=250)
#
## load the Doppler velocity data
#MIRA_VELg = larda.read("CLOUDNET", "v", [begin_dt, end_dt], [0, 'max'])
#LIMRAD94_VEL = larda.read("CLOUDNET_LIMRAD", "v", [begin_dt, end_dt], [0, 'max'])
#
#
#fig, ax = pyLARDA.Transformations.scatter(MIRA_VELg, LIMRAD94_VEL, var_lim=[-6, 4],
#                                          custom_offset_lines=1.0)
#
#fig.savefig(case_prefix+'cloudnet_scatter_mira_limrad_VEL.png', dpi=250)




__version__ = "3.3.2"
__author__ = "pyLARDA-dev-team"
__doc_link__ = "https://lacros-tropos.github.io/larda-doc/"

__init_text__ = f""">> LARDA initialized. Documentation available at {__doc_link__}"""

__default_info__ = """
The data from this campaign is provided by larda without warranty and liability.
Before publishing check the data license and contact the principal investigator.
Detailed information might be available using `larda.description('system', 'parameter')`."""



#!/usr/bin/python3

import toml
import numpy as np
import pprint
import collections
import logging
import re
logger = logging.getLogger(__name__)


def deep_update(source, overrides):
    """Update a nested dictionary.
    Only additions, no removal modify ``source`` in place.
    """
    for key, value in overrides.items():
        if isinstance(value, collections.Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
    return source


def select_matching_template(syskey, templates, fname):
    """templates either match by exact name or regex in also_match

    Example:
        in a template
        ```
        [CLOUDNET]
          also_match = "CLOUDNET_.+"
          [CLOUDNET.generic]
          ...
        ```
        also matches ``CLOUDNET_LIMRAD``
    """
    if syskey in templates:
        return templates[syskey]
    else:
        logger.info('no direct hit in template, try also_match')
        # (?=a)b regex will never match
        matches = [
            k for k, v in templates.items() 
            if bool(re.findall(v.get('also_match', '(?=a)b'), syskey))]
        assert len(matches) == 1, \
            f'more than one matching pattern found in template, check also_match tag in {fname}'
        return templates[matches[0]]


class ParameterInfo:
    """load the config file right here
    no need for prior allocation

    Args:
        config_path: location of the ``.toml`` config file
        config_file: name of the ``.toml`` config file
    """
    def __init__(self, config_path, config_file, cinfo_hand_down={}):
        logger.info("ParameterInfo: load config file {}".format(config_path, config_file))
        config = toml.load(config_path / config_file)
        self.config = {}
        
        for syskey, sysval in config.items():
            # get the template
            if 'template' in sysval:
                templates = toml.load(config_path / sysval['template'])
                temp = select_matching_template(syskey, templates, sysval['template'])
                #print('template', temp.keys())
                sysval = deep_update(temp, sysval)
            self.config[syskey] = sysval

        # do the inheritance of system level parameters here
        for syskey, sysval in self.config.items():
            #print("system level keys ", 
            #      [e for e in sysval.keys()])
            defaults = {**cinfo_hand_down, **sysval['generic']}
            #defaults = {k: v for k, v in sysval['generic'].items()}
            #pprint.pprint(defaults)
            for pkey, pval in sysval["params"].items():
                #print("param", pkey)
                #logger.debug(
                #    "paraminfo "+ pprint.pformat({**defaults, **pval, **{'system': syskey, 'name': pkey}}))               
                self.config[syskey]["params"][pkey] = {
                    **defaults, **pval, 
                    **{'system': syskey, 'paramkey': pkey}}
                if 'meta' in defaults and 'meta' in pval:
                    self.config[syskey]["params"][pkey]['meta'] = {
                        **defaults['meta'], **pval['meta']}
                
        

    def iterate_systems(self, keys=None):
        """provide iterator for the systems structure
        
        Args:
            keys: just iterate over the given (i.e. valid) systems
        
        Yields:
            syskey, config
        """

        if keys == None:
            keys = self.config.keys()
        else:
            keys = set(keys).intersection(self.config.keys())
        for syskey in keys:
            yield syskey, self.config[syskey]

    def name_and_unit(self):
                
        #if self.system_type=="Cloudnet":
            
        #    output = self.parameter_name
            
        #else :
        if self.unit == "" or self.unit == "[]":
            output = self.parameter_name
        else:
            if self.unit[0]=="[":
                output = self.parameter_name + " "+self.unit
            else:
                output = self.parameter_name + " ["+self.unit+"]"
        
        return output 
    
    def print_info(self):
                
        print("Parameter Name  : ",  self.parameter_name)
        print("System Type     : ",  self.system_type)
        print("Value Range Min : ",  self.value_range_min)
        print("Value Range Max : ",  self.value_range_max)
        print("COLORMAP        : ",  self.colormap)
        print("DIMENSIONS      : ",  self.dimensions)
        print("Variable  Name  : ",  self.variable_name)
        print("Time Var.       : ",  self.time_variable_name)
        print("Time Dim.       : ",  self.time_dimension_name)
        print("Range Dim.      : ",  self.range_dimension_name)
        print("Range Resolution: ",  self.range_resolution)
        print("Zero Bin        : ",  self.zero_bin)
        print("Zero Bin Height : ",  self.zero_bin_height)
        print("Unit            : ",  self.unit)
        print("Data Directory  : ",  self.data_dir)
        print("Filemask        : ",  self.filemask)
        print("Storage Type    : ",  self.storage_type)
        print("Display Type    : ",  self.display_type)
        print("Read Line       : ",  self.read_line)
        print("File Extension  : ",  self.file_extension)



#!/usr/bin/python3

"""

Try to use netcdf4-python

"""

import numpy as np
import netCDF4
import pyLARDA.helpers as h
from typing import List
import logging
import datetime
import scipy

logger = logging.getLogger(__name__)


def get_time_slicer(
    ts: np.array, 
    f: str, 
    time_interval: List[datetime.datetime]) -> list:
    """get time slicer from the time_interval
    Following options are available

    1. time_interval with [ts_begin, ts_end]
    2. only one timestamp is selected and the found right one would be beyond the ts range -> argnearest instead searchsorted
    3. only one is timestamp

    Args:
        ts: unix timestamps as array
        f: filename
        time_interval: time interval to slice for

    Returns:
        slicer


    """

    # select first timestamp right of begin (not left if nearer as above)
    #print(f'start time {h.ts_to_dt(ts[0])}')
    it_b = 0 if ts.shape[0] == 1 else np.searchsorted(ts, h.dt_to_ts(time_interval[0]), side='right')
    if len(time_interval) == 2:
        it_e = h.argnearest(ts, h.dt_to_ts(time_interval[1]))

        if it_b == ts.shape[0]: it_b = it_b - 1
        valid_step =  3 * np.median(np.diff(ts))
        if ts[it_e] < h.dt_to_ts(time_interval[0]) - valid_step or ts[it_b] < h.dt_to_ts(time_interval[0]):
            # second condition is to ensure that no timestamp before
            # the selected interval is choosen
            # (problem with limrad after change of sampling frequency)
            str = 'found last profile of file {}\n at ts[it_e] {} too far ({}s) from {}\n'.format(
                    f, h.ts_to_dt(ts[it_e]), valid_step, time_interval[0]) \
                 + 'or begin too early {} < {}\n returning None'.format(h.ts_to_dt(ts[it_b]), time_interval[0])
            logger.warning(str)
            return None

        it_e = it_e + 1 if not it_e == ts.shape[0] - 1 else None
        slicer = [slice(it_b, it_e)]
    elif it_b == ts.shape[0]:
        # only one timestamp is selected
        # and the found right one would be beyond the ts range
        it_b = h.argnearest(ts, h.dt_to_ts(time_interval[0]))
        slicer = [slice(it_b, it_b + 1)]
    else:
        slicer = [slice(it_b, it_b + 1)]
    return slicer


def get_var_attr_from_nc(name, paraminfo, variable):
    """get the attribute from the variable

    Args:
        name:
        paraminfo:
        variable:
    """

    direct_def = name.replace("identifier_", "")
    # if both are given (eg through inheritance, choose the
    # direct definition)
    logger.debug("attr name {}".format(name))
    attr = ''
    if name in paraminfo and direct_def not in paraminfo:
        try:
            attr = variable.getncattr(paraminfo[name])
        except Exception as e:
            logger.warning('Error extracting paraminfo of variable ' + str(name))
            logger.warning('Check spelling in .toml file or remove from .toml')
            logger.warning('Exception :: ', e)
    else:
        attr = paraminfo[name.replace("identifier_", "")]

    return attr


def get_meta_from_nc(ncD, meta_spec, varname):
    """get some meta data into the data_container

    specified within the paraminfo meta.name tags
    - gattr.name: global attribute with name
    - vattr.name: variable attribute with name
    - var.name: additional varable (ideally single value)

    Args:
        ncD: netCDF file object
        meta_spec: dict with all meta. definition
        varname: name of the variable to load (for vattr)

    Returns:
        dict with meta data
    """

    meta = {}
    for k, v in meta_spec.items():
        where, name = v.split('.')
        if where == 'gattr':
            meta[k] = [ncD.getncattr(name)]
        elif where == 'vattr':
            meta[k] = [ncD.variables[varname].getncattr(name)]
        elif where == 'var':
            meta[k] = [ncD.variables[name][:].data.tolist()]
        else:
            raise ValueError(f'meta string {v} for {k} not specified')
    return meta


def reader(paraminfo):
    """build a function for reading in time height data"""

    def retfunc(f, time_interval, *further_intervals):
        """function that converts the netCDF to the larda-data-format
        """
        logger.debug("filename at reader {}".format(f))
        with netCDF4.Dataset(f, 'r') as ncD:

            if 'auto_mask_scale' in paraminfo and paraminfo['auto_mask_scale'] == False:
                ncD.set_auto_mask(False)

            varconv_args = {}
            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec
            if 'base_time_variable' in paraminfo.keys() and \
                    paraminfo['base_time_variable'] in ncD.variables:
                basetime = ncD.variables[paraminfo['base_time_variable']][:].astype(np.float64)
                times += basetime

            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            if isinstance(times, np.ma.MaskedArray):
                ts = timeconverter(times.data)
            else:
                ts = timeconverter(times)
            # get the time slicer from time_interval
            slicer = get_time_slicer(ts, f, time_interval)
            if slicer is None and paraminfo['ncreader'] != 'pollynet_profile':
                logger.critical(f'No time slice found!\nfile :: {f}\n')
                return None

            if paraminfo['ncreader'] == "pollynet_profile":
                slicer = [slice(None)]

            if paraminfo['ncreader'] in ['timeheight', 'spec', 'mira_noise', 'pollynet_profile']:
                range_tg = True

                try:
                    range_interval = further_intervals[0]
                except IndexError as e:
                    logger.error('No range interval was given.')

                ranges = ncD.variables[paraminfo['range_variable']]
                logger.debug('loader range conversion {}'.format(paraminfo['range_conversion']))
                rangeconverter, _ = h.get_converter_array(
                    paraminfo['range_conversion'],
                    altitude=paraminfo['altitude'])
                ir_b = h.argnearest(rangeconverter(ranges[:]), range_interval[0])
                if len(range_interval) == 2:
                    if not range_interval[1] == 'max':
                        ir_e = h.argnearest(rangeconverter(ranges[:]), range_interval[1])
                        ir_e = ir_e + 1 if not ir_e == ranges.shape[0] - 1 else None
                    else:
                        ir_e = None
                    slicer.append(slice(ir_b, ir_e))
                else:
                    slicer.append(slice(ir_b, ir_b + 1))

            if paraminfo['ncreader'] == 'spec':
                if 'compute_velbins' in paraminfo:
                    if paraminfo['compute_velbins'] == "mrrpro":
                        wl = 1.238*10**(-2) # wavelength (fixed) - 24 GHz
                        varconv_args.update({"wl": wl})
                vel_tg = True
                slicer.append(slice(None))
            varconverter, maskconverter = h.get_converter_array(
                paraminfo['var_conversion'], **varconv_args)
            if 'vel_conversion' in paraminfo:
                velconverter, _ = h.get_converter_array(paraminfo['vel_conversion'])

            var = ncD.variables[paraminfo['variable_name']]
            # print('var dict ',ncD.variables[paraminfo['variable_name']].__dict__)
            # print("time indices ", it_b, it_e)
            data = {}
            if paraminfo['ncreader'] == 'timeheight':
                data['dimlabel'] = ['time', 'range']
            elif paraminfo['ncreader'] == 'time':
                data['dimlabel'] = ['time']
            elif paraminfo['ncreader'] == 'spec':
                data['dimlabel'] = ['time', 'range', 'vel']
            elif paraminfo['ncreader'] == 'mira_noise':
                data['dimlabel'] = ['time', 'range']
            elif paraminfo['ncreader'] == "pollynet_profile":
                data['dimlabel'] = ['time', 'range']

            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            if paraminfo['ncreader'] in ['timeheight', 'spec', 'mira_noise', 'pollynet_profile']:
                if isinstance(times, np.ma.MaskedArray):
                    data['rg'] = rangeconverter(ranges[tuple(slicer)[1]].data)
                else:
                    data['rg'] = rangeconverter(ranges[tuple(slicer)[1]])

                data['rg_unit'] = get_var_attr_from_nc("identifier_rg_unit",
                                                       paraminfo, ranges)
                logger.debug('shapes {} {} {}'.format(ts.shape, ranges.shape, var.shape))
            if paraminfo['ncreader'] == 'spec':
                if 'vel_ext_variable' in paraminfo:
                    # this special field is needed to load limrad spectra
                    vel_ext = ncD.variables[paraminfo['vel_ext_variable'][0]][int(paraminfo['vel_ext_variable'][1])]
                    vel_res = 2 * vel_ext / float(var[:].shape[2])
                    data['vel'] = np.linspace(-vel_ext + (0.5 * vel_res),
                                              +vel_ext - (0.5 * vel_res),
                                              var[:].shape[2])
                elif 'compute_velbins' in paraminfo:
                    if paraminfo['compute_velbins'] == 'mrrpro':
                    # this is used to read in MRR-PRO spectra
                        fs = 500000 # sampling rate of MRR-Pro (fixed)
                        vel_ext = fs/4/ncD.dimensions['range'].size*wl
                        vel_res = vel_ext / float(var[:].shape[2])
                        data['vel'] = np.linspace(0 - (0.5 * vel_res),
                                              -vel_ext + (0.5 * vel_res),
                                              var[:].shape[2])
                else:
                    data['vel'] = ncD.variables[paraminfo['vel_variable']][:]
                if 'vel_conversion' in paraminfo:
                    data['vel'] = velconverter(data['vel'])

            logger.debug('shapes {} {}'.format(ts.shape, var.shape))
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, var)]

            # by default assume dimensions of (time, range, ...)
            # or define a custom order in the param toml file
            if 'dimorder' in paraminfo:
                slicer = [slicer[i] for i in paraminfo['dimorder']]

            if paraminfo['ncreader'] == "pollynet_profile":
                del slicer[0]

            # read in the variable definition dictionary
            #
            if "identifier_var_def" in paraminfo.keys() and not "var_def" in paraminfo.keys():
                data['var_definition'] = h.guess_str_to_dict(
                    var.getncattr(paraminfo['identifier_var_def']))
            elif "var_def" in paraminfo.keys():
                data['var_definition'] =  paraminfo['var_def']

            if paraminfo['ncreader'] == 'mira_noise':
                r_c = ncD.variables[paraminfo['radar_const']][:]
                snr_c = ncD.variables[paraminfo['SNR_corr']][:]
                npw = ncD.variables[paraminfo['noise_pow']][:]
                calibrated_noise = r_c[slicer[0], np.newaxis] * var[tuple(slicer)].data * snr_c[tuple(slicer)].data / \
                                   npw[slicer[0], np.newaxis] * (data['rg'][np.newaxis, :] / 5000.) ** 2
                data['var'] = calibrated_noise
            else:
                data['var'] = varconverter(var[:])[tuple(slicer)]

                #if paraminfo['compute_velbins'] == "mrrpro":
                #    data['var'] = data['var'] * wl** 4 / (np.pi** 5) / 0.93 * 10**6

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                mask = np.isclose(data['var'].data, fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo['fill_value']
                mask = np.isclose(data['var'].data, fill_value)
            else:
                mask = ~np.isfinite(data['var'].data)
            
            #if isinstance(mask, np.ma.MaskedArray):
            #    mask = mask.mask
            assert not isinstance(mask, np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"
            data['mask'] = np.logical_or(mask, data['var'].mask)

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            if paraminfo['ncreader'] == "pollynet_profile":
                data['var'] = data['var'][np.newaxis, :]
                data['mask'] = data['mask'][np.newaxis, :]

            return data

    return retfunc


def auxreader(paraminfo):
    """build a function for reading in time height data"""

    def retfunc(f, time_interval, *further_intervals):
        """function that converts the netCDF to the larda-data-format
        this one is for aux values that don't have a dedicated time domain
        (nevertheless the time is read in, to estimate the coverage of the file)
        """
        logger.debug("filename at reader {}".format(f))
        with netCDF4.Dataset(f, 'r') as ncD:

            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec

            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            ts = timeconverter(times.data)

            # get the time slicer from time_interval
            slicer = get_time_slicer(ts, f, time_interval)
            if slicer == None and paraminfo['ncreader'] != 'aux_all_ts':
                return None

            if paraminfo['ncreader'] == "aux_all_ts":
                slicer = [slice(None)]

            varconverter, maskconverter = h.get_converter_array(
                paraminfo['var_conversion'])

            var = ncD.variables[paraminfo['variable_name']]
            # print('var dict ',ncD.variables[paraminfo['variable_name']].__dict__)
            # print("time indices ", it_b, it_e)
            data = {}
            data['dimlabel'] = ['time', 'aux']

            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[0:1]

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            logger.debug('shapes {} {}'.format(ts.shape, var.shape))
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, var)]

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                mask = (var[:] == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo['fill_value']
                mask = np.isclose(var[:], fill_value)
            else:
                mask = ~np.isfinite(var[:])

            if isinstance(mask, np.ma.MaskedArray):
                mask = mask.data
            assert not isinstance(mask, np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"

            data['var'] = varconverter(var[:])

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            data['mask'] = maskconverter(mask)

            return data

    return retfunc


def timeheightreader_rpgfmcw(paraminfo):
    """build a function for reading in time height data
    special function for a special instrument ;)

    the issues are:

    - range variable in different file
    - stacking of single variables

    for now works only with 3 chirps and range variable only in level0
    """

    def retfunc(f, time_interval, range_interval):
        """function that converts the netCDF to the larda-data-format
        """
        logger.debug("filename at reader {}".format(f))
        with netCDF4.Dataset(f, 'r') as ncD:

            no_chirps = ncD.dimensions['Chirp'].size

            ranges_per_chirp = [
                ncD.variables['C{}Range'.format(i + 1)] for i in range(no_chirps)]
            ch1range = ranges_per_chirp[0]

            ranges = np.hstack([rg[:] for rg in ranges_per_chirp])


            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec
            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            ts = timeconverter(times)

            # get the time slicer from time_interval
            slicer = get_time_slicer(ts, f, time_interval)
            if slicer == None:
                return None

            rangeconverter, _ = h.get_converter_array(
                paraminfo['range_conversion'])

            varconverter, _ = h.get_converter_array(
                paraminfo['var_conversion'])

            ir_b = h.argnearest(rangeconverter(ranges[:]), range_interval[0])
            if len(range_interval) == 2:
                if not range_interval[1] == 'max':
                    ir_e = h.argnearest(rangeconverter(ranges[:]), range_interval[1])
                    ir_e = ir_e + 1 if not ir_e == ranges.shape[0] - 1 else None
                else:
                    ir_e = None
                slicer.append(slice(ir_b, ir_e))
            else:
                slicer.append(slice(ir_b, ir_b + 1))

            no_chirps = ncD.dimensions['Chirp'].size

            var_per_chirp = [
                ncD.variables['C{}'.format(i + 1) + paraminfo['variable_name']] for i in range(no_chirps)]
            ch1var = var_per_chirp[0]

            # ch1var = ncD.variables['C1'+paraminfo['variable_name']]
            # ch2var = ncD.variables['C2'+paraminfo['variable_name']]
            # ch3var = ncD.variables['C3'+paraminfo['variable_name']]

            # print('var dict ',ch1var.__dict__)
            # print('shapes ', ts.shape, ch1range.shape, ch1var.shape)
            # print("time indices ", it_b, it_e)
            data = {}
            data['dimlabel'] = ['time', 'range']
            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]
            data['rg'] = rangeconverter(ranges[tuple(slicer)[1]])

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            data['rg_unit'] = get_var_attr_from_nc("identifier_rg_unit",
                                                   paraminfo, ch1range)
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, ch1var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, ch1var)]
            var = np.hstack([v[:] for v in var_per_chirp])
            # var = np.hstack([ch1var[:], ch2var[:], ch3var[:]])

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                data['mask'] = (var[tuple(slicer)].data == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo["fill_value"]
                data['mask'] = np.isclose(var[tuple(slicer)].data, fill_value)
            else:
                data['mask'] = ~np.isfinite(var[tuple(slicer)].data)

            assert not isinstance(data['mask'], np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"
            data['var'] = varconverter(var[tuple(slicer)].data)

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            return data

    return retfunc


def specreader_rpgfmcw(paraminfo):
    """build a function for reading in spectral data
    special function for a special instrument ;)

    the issues are:

    - range variable in different file
    - stacking of single variables

    for now works only with 3 chirps and range variable only in level0
    """

    def retfunc(f, time_interval, range_interval):
        """function that converts the netCDF to the larda-data-format
        """
        logger.debug("filename at reader {}".format(f))

        with netCDF4.Dataset(f, 'r') as ncD:

            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec
            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            ts = timeconverter(times)

            no_chirps = ncD.dimensions['Chirp'].size

            ranges_per_chirp = [
                ncD.variables['C{}Range'.format(i + 1)] for i in range(no_chirps)]
            ch1range = ranges_per_chirp[0]

            ranges = np.hstack([rg[:] for rg in ranges_per_chirp])

            # get the time slicer from time_interval
            slicer = get_time_slicer(ts, f, time_interval)
            if slicer == None:
                return None

            rangeconverter, _ = h.get_converter_array(
                paraminfo['range_conversion'])

            varconverter, _ = h.get_converter_array(
                paraminfo['var_conversion'])

            ir_b = h.argnearest(rangeconverter(ranges[:]), range_interval[0])
            if len(range_interval) == 2:
                if not range_interval[1] == 'max':
                    ir_e = h.argnearest(rangeconverter(ranges[:]), range_interval[1])
                    ir_e = ir_e + 1 if not ir_e == ranges.shape[0] - 1 else None
                else:
                    ir_e = None
                slicer.append(slice(ir_b, ir_e))
            else:
                slicer.append(slice(ir_b, ir_b + 1))

            vars_per_chirp = [
                ncD.variables['C{}{}'.format(i + 1, paraminfo['variable_name'])] for i in range(no_chirps)]
            ch1var = vars_per_chirp[0]
            # print('var dict ',ch1var.__dict__)
            # print('shapes ', ts.shape, ch1range.shape, ch1var.shape)
            # print("time indices ", it_b, it_e)

            data = {}
            data['dimlabel'] = ['time', 'range', 'vel']
            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]
            data['rg'] = rangeconverter(ranges[tuple(slicer)[1]])

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            data['rg_unit'] = get_var_attr_from_nc("identifier_rg_unit",
                                                   paraminfo, ch1range)
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, ch1var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, ch1var)]
            if 'vel_ext_variable' in paraminfo:
                # define the function
                get_vel_ext = lambda i: ncD.variables[paraminfo['vel_ext_variable'][0]][:][i]
                # apply it to every chirp
                vel_ext_per_chirp = [get_vel_ext(i) for i in range(no_chirps)]

                vel_dim_per_chirp = [v.shape[2] for v in vars_per_chirp]
                calc_vel_res = lambda v_e, v_dim: 2.0 * v_e / float(v_dim)
                vel_res_per_chirp = [calc_vel_res(v_e, v_dim) for v_e, v_dim \
                                     in zip(vel_ext_per_chirp, vel_dim_per_chirp)]

                # for some very obscure reason lambda is not able to unpack 3 values
                def calc_vel(vel_ext, vel_res, v_dim):
                    return np.linspace(-vel_ext + (0.5 * vel_res),
                                       +vel_ext - (0.5 * vel_res),
                                       v_dim)

                vel_per_chirp = [calc_vel(v_e, v_res, v_dim) for v_e, v_res, v_dim \
                                 in zip(vel_ext_per_chirp, vel_res_per_chirp, vel_dim_per_chirp)]
            else:
                raise NotImplemented("other means of getting the var dimension are not implemented yet")
            data['vel'] = vel_per_chirp[0]

            # interpolate the variables here
            if 'var_conversion' in paraminfo and paraminfo['var_conversion'] == 'keepNyquist':
                # the interpolation is only done for the number of spectral lines, not the velocity itself
                quot = [i/vel_dim_per_chirp[0] for i in vel_dim_per_chirp[1:]]
                vars_interp = [vars_per_chirp[0]]
                ich = 1
                for var, vel in zip(vars_per_chirp[1:], vel_per_chirp[1:]):
                    data['vel_ch{}'.format(ich+1)] = vel_per_chirp[ich]
                    new_vel = np.linspace(vel[0], vel[-1], vel_dim_per_chirp[0])
                    vars_interp.append(interp_only_3rd_dim(var[:] * quot[ich-1], vel, new_vel, kind='nearest'))
                    ich += 1
            else:
                vars_interp = [vars_per_chirp[0]] + \
                              [interp_only_3rd_dim(var, vel, vel_per_chirp[0]) \
                               for var, vel in zip(vars_per_chirp[1:], vel_per_chirp[1:])]


            var = np.hstack([v[:] for v in vars_interp])
            logger.debug('interpolated spectra from\n{}\n{} to\n{}'.format(
                [v[:].shape for v in vars_per_chirp],
                ['{:5.3f}'.format(vel[0]) for vel in vel_per_chirp],
                [v[:].shape for v in vars_interp]))
            logger.info('var.shape interpolated spectra {}'.format(var.shape))

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                data['mask'] = (var[tuple(slicer)].data == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo["fill_value"]
                data['mask'] = np.isclose(var[tuple(slicer)].data, fill_value)
            else:
                data['mask'] = ~np.isfinite(var[tuple(slicer)].data)
            if isinstance(times, np.ma.MaskedArray):
                data['var'] = varconverter(var[tuple(slicer)].data)
            else:
                data['var'] = varconverter(var[tuple(slicer)].data)


            assert not isinstance(data['mask'], np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            return data

    return retfunc


def specreader_rpgpy(paraminfo):
    """build a function for reading in spectral data
    """

    def retfunc(f, time_interval, range_interval):
        """function that converts the netCDF to the larda-data-format
        """
        logger.debug("filename at reader {}".format(f))

        with netCDF4.Dataset(f, 'r') as ncD:

            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec
            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            ts = timeconverter(times)

            no_chirps = ncD.dimensions['chirp'].size

            # check if spectra will be interpolated over chirps or if we extract only one chirp
            interpolate_velocity = paraminfo['paramkey'][0] != 'C'
            if not interpolate_velocity:
                print(f'Key {paraminfo["paramkey"]} starts with "C", only reading chirp {paraminfo["paramkey"][1]}. ')
            ranges = ncD.variables[paraminfo['range_variable']]

            # get the time slicer from time_interval
            slicer = get_time_slicer(ts, f, time_interval)
            if slicer == None:
                return None

            rangeconverter, _ = h.get_converter_array(
                paraminfo['range_conversion'])

            varconverter, _ = h.get_converter_array(
                paraminfo['var_conversion'])

            ir_b = h.argnearest(rangeconverter(ranges[:]), range_interval[0])
            if len(range_interval) == 2:
                if not range_interval[1] == 'max':
                    ir_e = h.argnearest(rangeconverter(ranges[:]), range_interval[1])
                    ir_e = ir_e + 1 if not ir_e == ranges.shape[0] - 1 else None
                else:
                    ir_e = None
                slicer.append(slice(ir_b, ir_e))
            else:
                slicer.append(slice(ir_b, ir_b + 1))

            var = ncD.variables[paraminfo['variable_name']]

            data = {}
            data['dimlabel'] = ['time', 'range', 'vel']
            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]
            data['rg'] = rangeconverter(ranges[tuple(slicer)[1]])

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            data['rg_unit'] = get_var_attr_from_nc("identifier_rg_unit",
                                                   paraminfo, ranges)
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, var)]
            if no_chirps > 1:
                range_offsets = np.hstack((ncD.variables['chirp_start_indices'][:], ncD.variables['n_range_layers'][:]))
                if interpolate_velocity:
                    common_velocity = np.linspace(-np.nanmax(ncD.variables['velocity_vectors']),
                                                  np.nanmax(ncD.variables['velocity_vectors']),
                                                  ncD.variables['velocity_vectors'][0].shape[0])
                if interpolate_velocity:
                    var_array = []
                    for i in range(no_chirps):
                        v = ncD.variables['velocity_vectors'][:][i]
                        valid_indices = v != -999
                        dv = np.nanmean(np.diff(v[valid_indices]))
                        var_per_chirp = var[tuple(slicer)[0], range_offsets[i]: range_offsets[i+1], valid_indices] / dv
                        f = scipy.interpolate.interp1d(v[valid_indices], var_per_chirp, bounds_error=False, fill_value=0.)
                        v_interp = f(common_velocity)
                        var_array.append(v_interp)

                    var_array = np.hstack(var_array)
                    dv2 = common_velocity[5] - common_velocity[4]
                    var_out = var_array*dv2

                    data['vel'] = common_velocity
                else:
                    chirp_to_extract = int(paraminfo['paramkey'][1])
                    assert(chirp_to_extract <= no_chirps), f"chirp to extract is {chirp_to_extract} but number of chirps" \
                                                           f" is {no_chirps}."
                    var_out = np.zeros(var.shape)
                    v = ncD.variables['velocity_vectors'][:][chirp_to_extract-1]
                    valid_indices = ~v.mask
                    var_out[tuple(slicer)[0], range_offsets[chirp_to_extract-1]: range_offsets[chirp_to_extract], :] = \
                        var[tuple(slicer)[0], range_offsets[chirp_to_extract-1]: range_offsets[chirp_to_extract], :]

                    var_out = var_out[tuple(slicer)[0], :, valid_indices]
                    data['vel'] = v[valid_indices]

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                data['mask'] = (var_out[:, tuple(slicer)[1]] == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo["fill_value"]
                data['mask'] = np.isclose(var_out[:, tuple(slicer)[1]], fill_value)
            else:
                data['mask'] = ~np.isfinite(var_out[:, tuple(slicer)[1]])
            if isinstance(times, np.ma.MaskedArray):
                data['var'] = varconverter(var_out[:, tuple(slicer)[1]])
            else:
                data['var'] = varconverter(var_out[:, tuple(slicer)[1]])

            assert not isinstance(data['mask'], np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            return data

    return retfunc


def scanreader_mira(paraminfo):
    """reader for the scan files

    - load full file regardless of selected time
    - covers spec_timeheight and spec_time
    """

    def retfunc(f, time_interval, *further_intervals):
        """function that converts the netCDF to the larda-data-format
        """
        logger.debug("filename at reader {}".format(f))
        with netCDF4.Dataset(f, 'r') as ncD:

            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec

            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            if isinstance(times, np.ma.MaskedArray):
                ts = timeconverter(times.data)
            else:
                ts = timeconverter(times)

            # load the whole time-range from the file
            slicer = [slice(None)]

            if paraminfo['ncreader'] == 'scan_timeheight':
                range_tg = True

                range_interval = further_intervals[0]
                ranges = ncD.variables[paraminfo['range_variable']]
                logger.debug('loader range conversion {}'.format(paraminfo['range_conversion']))
                rangeconverter, _ = h.get_converter_array(
                    paraminfo['range_conversion'],
                    altitude=paraminfo['altitude'])
                ir_b = h.argnearest(rangeconverter(ranges[:]), range_interval[0])
                if len(range_interval) == 2:
                    if not range_interval[1] == 'max':
                        ir_e = h.argnearest(rangeconverter(ranges[:]), range_interval[1])
                        ir_e = ir_e + 1 if not ir_e == ranges.shape[0] - 1 else None
                    else:
                        ir_e = None
                    slicer.append(slice(ir_b, ir_e))
                else:
                    slicer.append(slice(ir_b, ir_b + 1))

            varconverter, maskconverter = h.get_converter_array(
                paraminfo['var_conversion'],
                mira_azi_zero=paraminfo['mira_azi_zero'])

            var = ncD.variables[paraminfo['variable_name']]
            # print('var dict ',ncD.variables[paraminfo['variable_name']].__dict__)
            # print("time indices ", it_b, it_e)
            data = {}
            if paraminfo['ncreader'] == 'scan_timeheight':
                data['dimlabel'] = ['time', 'range']
            elif paraminfo['ncreader'] == 'scan_time':
                data['dimlabel'] = ['time']
            # elif paraminfo['ncreader'] == 'spec':
            #    data['dimlabel'] = ['time', 'range', 'vel']

            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            if paraminfo['ncreader'] == 'scan_timeheight':
                if isinstance(times, np.ma.MaskedArray):
                    data['rg'] = rangeconverter(ranges[tuple(slicer)[1]].data)
                else:
                    data['rg'] = rangeconverter(ranges[tuple(slicer)[1]])

                data['rg_unit'] = get_var_attr_from_nc("identifier_rg_unit",
                                                       paraminfo, ranges)
                logger.debug('shapes {} {} {}'.format(ts.shape, ranges.shape, var.shape))
            logger.debug('shapes {} {}'.format(ts.shape, var.shape))
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, var)]

            # by default assume dimensions of (time, range, ...)
            # or define a custom order in the param toml file
            if 'dimorder' in paraminfo:
                slicer = [slicer[i] for i in paraminfo['dimorder']]

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                mask = (var[tuple(slicer)].data == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo['fill_value']
                mask = np.isclose(var[tuple(slicer)].data, fill_value)
            else:
                mask = ~np.isfinite(var[tuple(slicer)].data)

            assert not isinstance(mask, np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"

            data['var'] = varconverter(var[tuple(slicer)].data)

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            data['mask'] = maskconverter(mask)

            return data

    return retfunc


def interp_only_3rd_dim(arr, old, new, **kwargs):
    """function to interpolate only the velocity (3rd) axis"""

    from scipy import interpolate

    kind_ = kwargs['kind'] if 'kind' in kwargs else 'linear'

    f = interpolate.interp1d(old, arr, axis=2,
                             bounds_error=False, fill_value=-999., kind=kind_)
    new_arr = f(new)

    return new_arr


def specreader_kazr(paraminfo):
    """build a function for reading in spectral data
    another special function for another special instrument ;)

    the issues are:

    - variables time and range are merged and can be accessed by a locator mask
    - noise is not saved and has to be computed from the spectra
    - spectra are not in reflectivity but in 10*log10(mW)
    - need to be converted using specZ = spec * h.z2lin(float(f.cal_constant[:-3])) * self.range[ir]**2

    """

    def retfunc(f, time_interval, range_interval):
        """function that converts the netCDF to the larda-data-format
        """
        logger.debug("filename at reader {}".format(f))

        with netCDF4.Dataset(f, 'r') as ncD:
            ranges = ncD.variables[paraminfo['range_variable']]
            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)
            locator_mask = ncD.variables[paraminfo['mask_var']][:].astype(np.int)
            if 'time_millisec_variable' in paraminfo.keys() and \
                    paraminfo['time_millisec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_millisec_variable']][:] / 1.0e3
                times += subsec
            if 'time_microsec_variable' in paraminfo.keys() and \
                    paraminfo['time_microsec_variable'] in ncD.variables:
                subsec = ncD.variables[paraminfo['time_microsec_variable']][:] / 1.0e6
                times += subsec
            if 'base_time_variable' in paraminfo.keys() and \
                    paraminfo['base_time_variable'] in ncD.variables:
                basetime = ncD.variables[paraminfo['base_time_variable']][:].astype(np.float64)
                times += basetime
            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            ts = timeconverter(times)

            it_b = np.searchsorted(ts, h.dt_to_ts(time_interval[0]), side='right')
            if len(time_interval) == 2:
                it_e = h.argnearest(ts, h.dt_to_ts(time_interval[1]))
                if it_b == ts.shape[0]: it_b = it_b - 1
                if ts[it_e] < h.dt_to_ts(time_interval[0]) - 3 * np.median(np.diff(ts)) \
                        or ts[it_b] < h.dt_to_ts(time_interval[0]):
                    # second condition is to ensure that no timestamp before
                    # the selected interval is chosen
                    # (problem with limrad after change of sampling frequency)
                    logger.warning(
                        'last profile of file {}\n at {} too far from {}'.format(
                            f, h.ts_to_dt(ts[it_e]), time_interval[0]))
                    return None
                it_e = it_e + 1 if not it_e == ts.shape[0] - 1 else None
                slicer = [slice(it_b, it_e)]
            elif it_b == ts.shape[0]:
                # only one timestamp is selected
                # and the found right one would be beyond the ts range
                it_b = h.argnearest(ts, h.dt_to_ts(time_interval[0]))
                slicer = [slice(it_b, it_b + 1)]
            else:
                slicer = [slice(it_b, it_b + 1)]

            rangeconverter, _ = h.get_converter_array(
                paraminfo['range_conversion'])

            varconverter, _ = h.get_converter_array(
                paraminfo['var_conversion'])

            ir_b = h.argnearest(rangeconverter(ranges[:]), range_interval[0])
            if len(range_interval) == 2:
                if not range_interval[1] == 'max':
                    ir_e = h.argnearest(rangeconverter(ranges[:]), range_interval[1])
                    ir_e = ir_e + 1 if not ir_e == ranges.shape[0] - 1 else None
                else:
                    ir_e = None
                slicer.append(slice(ir_b, ir_e))
            else:
                slicer.append(slice(ir_b, ir_b + 1))

            range_out = rangeconverter(ranges[tuple(slicer)[1]])
            cal = getattr(ncD, paraminfo['cal_const'])
            var = ncD.variables[paraminfo['variable_name']][:].astype(np.float64)
            var = var[locator_mask]
            vel = ncD.variables[paraminfo['vel_variable']][:].astype(np.float64)
            # print('var dict ',ch1var.__dict__)
            # print('shapes ', ts.shape, ch1range.shape, ch1var.shape)
            # print("time indices ", it_b, it_e)

            data = {}
            data['dimlabel'] = ['time', 'range', 'vel']
            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]
            data['rg'] = range_out

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            data['rg_unit'] = get_var_attr_from_nc("identifier_rg_unit",
                                                   paraminfo, ranges)
            #data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
            #                                        paraminfo, var)
            data['var_unit'] = 'dBZ m-1 s'
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, var)]
            data['vel'] = vel

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                data['mask'] = (var[tuple(slicer)].data == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo["fill_value"]
                data['mask'] = np.isclose(var[tuple(slicer)], fill_value)
            elif "mask_var" in paraminfo.keys():
                # combine locator mask and mask of infinite values
                mask = locator_mask.mask[tuple(slicer)]
                data["mask"] = np.logical_or(~np.isfinite(var[tuple(slicer)].data), np.repeat(mask[:,:,np.newaxis],len(data['vel']), axis=2))
            else:
                data['mask'] = ~np.isfinite(var[tuple(slicer)].data)
            if isinstance(times, np.ma.MaskedArray):
                var = varconverter(var[tuple(slicer)].data)
            else:
                var = varconverter(var[tuple(slicer)])

            assert not isinstance(mask, np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"

            var2 = h.z2lin(var) * h.z2lin(float(cal[:-3])) * (range_out ** 2)[np.newaxis, :, np.newaxis]
            data['var'] = var2

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            return data

    return retfunc


def reader_pollyraw(paraminfo):
    """build a function for reading in the polly raw data into larda"""

    def retfunc(f, time_interval, *further_intervals):
        """function that converts the netCDF to the larda container
        """
        logger.debug("filename at reader {}".format(f))
        import zipfile
        import os
        with zipfile.ZipFile(f) as zfile:
            path, file = os.path.split(f)
            ncD = netCDF4.Dataset('dummy', mode='r',
                                  memory=zfile.read(file[:-4]))

            times = ncD.variables[paraminfo['time_variable']][:].astype(np.float64)

            timeconverter, _ = h.get_converter_array(
                paraminfo['time_conversion'], ncD=ncD)
            if isinstance(times, np.ma.MaskedArray):
                ts = timeconverter(times.data)
            else:
                ts = timeconverter(times)
            # get the time slicer from time_interval
            slicer = get_time_slicer(ts, f, time_interval)
            if slicer == None:
                return None

            # load just the first 2500 range bins of polly
            slicer.append(slice(0, 2500))
            varconverter, maskconverter = h.get_converter_array(
                paraminfo['var_conversion'])

            varname, dim = paraminfo['variable_name'].split(':')
            slicer.append(int(dim))
            var = ncD.variables[varname]
            # print('var dict ',ncD.variables[paraminfo['variable_name']].__dict__)
            # print("time indices ", it_b, it_e)
            data = {}
            data['dimlabel'] = ['time', 'range']

            data["filename"] = f
            data["paraminfo"] = paraminfo
            data['ts'] = ts[tuple(slicer)[0]]

            data['system'] = paraminfo['system']
            data['name'] = paraminfo['paramkey']
            data['colormap'] = paraminfo['colormap']

            if 'meta' in paraminfo:
                data['meta'] = get_meta_from_nc(ncD, paraminfo['meta'], paraminfo['variable_name'])

            # also experimental: vis_varconverter
            if 'plot_varconverter' in paraminfo and paraminfo['plot_varconverter'] != 'none':
                data['plot_varconverter'] = paraminfo['plot_varconverter']
            else:
                data['plot_varconverter'] = ''

            data['rg'] = np.arange(0, 2500)

            data['rg_unit'] = 'range_bin'
            logger.debug('shapes {} {}'.format(ts.shape, var.shape))
            data['var_unit'] = get_var_attr_from_nc("identifier_var_unit",
                                                    paraminfo, var)
            data['var_lims'] = [float(e) for e in \
                                get_var_attr_from_nc("identifier_var_lims",
                                                     paraminfo, var)]

            # by default assume dimensions of (time, range, ...)
            # or define a custom order in the param toml file
            if 'dimorder' in paraminfo:
                slicer = [slicer[i] for i in paraminfo['dimorder']]

            if "identifier_fill_value" in paraminfo.keys() and not "fill_value" in paraminfo.keys():
                fill_value = var.getncattr(paraminfo['identifier_fill_value'])
                mask = (var[tuple(slicer)].data == fill_value)
            elif "fill_value" in paraminfo.keys():
                fill_value = paraminfo['fill_value']
                mask = np.isclose(var[tuple(slicer)].data, fill_value)
            else:
                mask = ~np.isfinite(var[tuple(slicer)].data)

            assert not isinstance(mask, np.ma.MaskedArray), \
               "mask array shall not be np.ma.MaskedArray, but of plain booltype"

            #print(slicer)
            data['var'] = varconverter(var[tuple(slicer)].data)

            if isinstance(data['var'], np.ma.MaskedArray):
                data['var'] = data['var'].data
            assert not isinstance(data['var'], np.ma.MaskedArray), \
               "var array shall not be np.ma.MaskedArray, but of plain booltype"

            data['mask'] = maskconverter(mask)

            return data

    return retfunc


def reader_wyoming_sounding(paraminfo):
    """
    build a reader to read in Wyoming Upper Air soundings, saved locally as a txt file
    Args:
        paraminfo: parameter information from toml file

    Returns:
        reader function

    """
    def retfunc(f, time_interval, *further_intervals):
        """
        function that converts the txt file to larda data container
        Args:
            f:
            time_interval:

        Returns:
            larda data container with sounding data
        """
        import csv
        import datetime
        logger.debug("filename at reader {}".format(f))
        with open(f) as f:
            reader = csv.reader(f, delimiter='\t')
            headers = next(reader, None)
            var_index = [i for i,j in enumerate(headers) if j == paraminfo['variable_name']]
            assert(len(var_index) == 1), "mismatch between headers in file and variable name in toml"
            rg_index = [i for i,j in enumerate(headers) if j == paraminfo['range_variable']]
            assert(len(rg_index) == 1), "mismatch between headers in file and range variable name in toml"
            data = {}
            data['dimlabel'] = ['time', 'range']
            data['ts'] = np.array([h.dt_to_ts(datetime.datetime.strptime(f.name.split('/')[-1][0:11], '%Y%m%d_%H'))])
            data['var'] = []
            data['rg'] = []
            for row in reader:
                try:
                    data['var'].append(float(row[var_index[0]]))
                except ValueError:  # empty line cannot be converted to float
                    data['var'].append(np.nan)
                data['rg'].append(float(row[rg_index[0]]))
            data['var'] = np.array(data['var'])[np.newaxis,:]
            data['rg'] = np.array(data['rg'])
            data['mask'] = np.isnan(data['var'])
            data['name'] = paraminfo['paramkey']
            data['system'] = paraminfo['system']
            data['var_lims'] = paraminfo['var_lims']
            data['colormap'] = 'jet' if not 'colormap' in paraminfo else paraminfo['colormap']
            data['rg_unit'] = paraminfo['rg_unit']
            data['var_unit'] = paraminfo['var_unit']
            data['paraminfo'] = paraminfo
            data['filename'] = f.name
            return data

    return retfunc



#!/usr/bin/python3

import sys
# just needed to find pyLARDA from this location
sys.path.append('../')
sys.path.append('.')

import matplotlib
matplotlib.use('Agg')
import pyLARDA
import pyLARDA.helpers as h
import datetime
import scipy.ndimage as spn

import logging
log = logging.getLogger('pyLARDA')
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler())

#Load LARDA
larda=pyLARDA.LARDA().connect('lacros_dacapo')
c_info = [larda.camp.LOCATION, larda.camp.VALID_DATES]

print('available systems:', larda.connectors.keys())
print("available parameters: ",[(k, larda.connectors[k].params_list) for k in larda.connectors.keys()])
print('days with data', larda.days_with_data())
#print("array_avail()", larda.array_avail(2015, 6))
#print("single month with new interface ", larda.instr_status(2015, 6)) 

#begin_dt=datetime.datetime(2018,12,2,16,0,0)
#end_dt=datetime.datetime(2018,12,2,22,0,0)

begin_dt=datetime.datetime(2018,12,18,6,0)
end_dt=datetime.datetime(2018,12,18,11,0,0)
plot_range = [5500, 7500]


MIRA_Zg=larda.read("MIRA","Zg",[begin_dt,end_dt],[0,'max'])
#fig, ax = pyLARDA.Transformations.plot_timeheight(params["MIRA_Zg"],begin,end,plot_base,plot_top,savename="example/mira_zg.png")
# customly overwrite mira plot limits
MIRA_Zg['var_lims'] = [-40,20]
fig, ax = pyLARDA.Transformations.plot_timeheight(MIRA_Zg, range_interval=plot_range, z_converter='lin2z')
fig.savefig('mira_z.png', dpi=250)
h.pprint(MIRA_Zg)

LIMRAD94_Z=larda.read("LIMRAD94","Ze",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(LIMRAD94_Z, range_interval=plot_range, z_converter='lin2z')
fig.savefig('limrad_Z.png', dpi=250)

MRR_Z=larda.read("MRRPRO","Ze",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(MRR_Z, range_interval=plot_range, z_converter='lin2z')
fig.savefig('mrr_Z.png', dpi=250)

shaun_vel=larda.read("SHAUN","VEL",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(shaun_vel, range_interval=plot_range)
fig.savefig('shaun_vel.png', dpi=250)


cloudnet_beta=larda.read("CLOUDNET","beta",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(cloudnet_beta, range_interval=plot_range, z_converter="log")
fig.savefig('cloudnet_beta.png', dpi=250)

cloudnet_depol=larda.read("CLOUDNET","depol",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(cloudnet_depol, range_interval=plot_range)
fig.savefig('cloudnet_delta.png', dpi=250)

CLOUDNET_lwp=larda.read("CLOUDNET", "LWP", [begin_dt,end_dt])
fig, ax = pyLARDA.Transformations.plot_timeseries(CLOUDNET_lwp)
fig.savefig('cloudnet_lwp.png', dpi=250)

CLOUDNET_Z=larda.read("CLOUDNET","Z",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(CLOUDNET_Z, range_interval=plot_range, z_converter='lin2z')
fig.savefig('cloudnet_Z.png', dpi=250)

CLOUDNET_VEL=larda.read("CLOUDNET","VEL",[begin_dt,end_dt],[0,'max'])
fig, ax = pyLARDA.Transformations.plot_timeheight(CLOUDNET_VEL, range_interval=plot_range)
fig.savefig('cloudnet_VEL.png', dpi=250)
print("Simple Example - END")
